"""
This module provides functions for parallel execution of loops and function
mappings, using the builtin Python module multiprocessing or the loky parallel
execution library.
"""
__all__ = ['parallel_map', 'serial_map', 'loky_pmap', 'mpi_pmap']

import multiprocessing
import os
import sys
import time
import threading
import concurrent.futures
import warnings
from qutip.ui.progressbar import progress_bars
from qutip.settings import available_cpu_count

if sys.platform == 'darwin':
    mp_context = multiprocessing.get_context('fork')
elif sys.platform == 'linux':
    # forkserver would handle threads better, but is much slower at starting
    # the executor and spawning tasks
    mp_context = multiprocessing.get_context('fork')
else:
    mp_context = multiprocessing.get_context()


default_map_kw = {
    'timeout': threading.TIMEOUT_MAX,
    'num_cpus': available_cpu_count(),
    'fail_fast': True,
}


def _read_map_kw(options):
    options = options or {}
    map_kw = default_map_kw.copy()
    map_kw.update({k: v for k, v in options.items() if v is not None})
    return map_kw


class MapExceptions(Exception):
    def __init__(self, msg, errors, results):
        super().__init__(msg, errors, results)
        self.errors = errors
        self.results = results


def serial_map(task, values, task_args=None, task_kwargs=None,
               reduce_func=None, map_kw=None,
               progress_bar=None, progress_bar_kwargs={}):
    """
    Serial mapping function with the same call signature as parallel_map, for
    easy switching between serial and parallel execution. This
    is functionally equivalent to::

        result = [task(value, *task_args, **task_kwargs) for value in values]

    This function work as a drop-in replacement of :func:`parallel_map`.

    Parameters
    ----------
    task : a Python function
        The function that is to be called for each value in ``task_vec``.
    values : array / list
        The list or array of values for which the ``task`` function is to be
        evaluated.
    task_args : list, optional
        The optional additional argument to the ``task`` function.
    task_kwargs : dictionary, optional
        The optional additional keyword argument to the ``task`` function.
    reduce_func : func, optional
        If provided, it will be called with the output of each tasks instead of
        storing a them in a list. It should return None or a number.
        When returning a number, it represent the estimation of the number of
        task left. On a return <= 0, the map will end early.
    progress_bar : str, optional
        Progress bar options's string for showing progress.
    progress_bar_kwargs : dict, optional
        Options for the progress bar.
    map_kw: dict, optional
        Dictionary containing:
        - timeout: float, Maximum time (sec) for the whole map.
        - fail_fast: bool, Raise an error at the first.

    Returns
    -------
    result : list
        The result list contains the value of
        ``task(value, *task_args, **task_kwargs)`` for each
        value in ``values``. If a ``reduce_func`` is provided, and empty list
        will be returned.

    """
    if task_args is None:
        task_args = ()
    if task_kwargs is None:
        task_kwargs = {}
    map_kw = _read_map_kw(map_kw)
    remaining_ntraj = None
    progress_bar = progress_bars[progress_bar](
        len(values), **progress_bar_kwargs
    )
    end_time = map_kw['timeout'] + time.time()
    results = None
    if reduce_func is None:
        results = [None] * len(values)
    errors = {}
    for n, value in enumerate(values):
        if time.time() > end_time:
            break
        progress_bar.update()
        try:
            result = task(value, *task_args, **task_kwargs)
        except Exception as err:
            if map_kw["fail_fast"]:
                raise err
            else:
                errors[n] = err
        else:
            if reduce_func is not None:
                remaining_ntraj = reduce_func(result)
            else:
                results[n] = result
        if remaining_ntraj is not None and remaining_ntraj <= 0:
            end_time = 0
    progress_bar.finished()

    if errors:
        raise MapExceptions(f"{len(errors)} iterations failed in serial_map",
                            errors, results)
    return results


def _generic_pmap(task, values, task_args, task_kwargs, reduce_func,
                  timeout, fail_fast, num_workers,
                  progress_bar, progress_bar_kwargs,
                  setup_executor, extract_result, shutdown_executor):
    """
    Common functionality for parallel_map, loky_pmap and mpi_pmap.
    The parameters `setup_executor`, `extract_result` and `shutdown_executor`
    are callback functions with the following signatures:

    setup_executor: () -> ProcessPoolExecutor

    extract_result: Future -> (Any, BaseException)
        If there was an exception e, returns (None, e).
        Otherwise returns (result, None).

    shutdown_executor: (executor: ProcessPoolExecutor,
                        active_tasks: set[Future]) -> None
        executor: The ProcessPoolExecutor that was created in setup_executor
        active_tasks: A set of Futures that are currently still being executed
            (non-empty if: timeout, error, or reduce_func requesting exit)
    """

    if task_args is None:
        task_args = ()
    if task_kwargs is None:
        task_kwargs = {}
    end_time = timeout + time.time()

    progress_bar = progress_bars[progress_bar](
        len(values), **progress_bar_kwargs
    )

    errors = {}
    finished = []
    if reduce_func is not None:
        results = None

        def result_func(_, value):
            return reduce_func(value)
    else:
        results = [None] * len(values)
        result_func = results.__setitem__

    def _done_callback(future):
        if not future.cancelled():
            result, exception = extract_result(future)
            if isinstance(exception, KeyboardInterrupt):
                # When a keyboard interrupt happens, it is raised in the main
                # thread and in all worker threads. At this point in the code,
                # the worker threads have already returned and the main thread
                # is only waiting for the ProcessPoolExecutor to shutdown
                # before exiting. We therefore return immediately.
                return
            if exception is not None:
                if isinstance(exception, Exception):
                    errors[future._i] = exception
                else:
                    raise exception
            else:
                remaining_ntraj = result_func(future._i, result)
                if remaining_ntraj is not None and remaining_ntraj <= 0:
                    finished.append(True)
        progress_bar.update()

    os.environ['QUTIP_IN_PARALLEL'] = 'TRUE'
    try:
        with setup_executor() as executor:
            waiting = set()
            i = 0
            aborted = False

            while i < len(values):
                # feed values to the executor, ensuring that there is at
                # most one task per worker at any moment in time so that
                # we can shutdown without waiting for greater than the time
                # taken by the longest task
                if len(waiting) >= num_workers:
                    # no space left, wait for a task to complete or
                    # the time to run out
                    timeout = max(0, end_time - time.time())
                    _done, waiting = concurrent.futures.wait(
                        waiting,
                        timeout=timeout,
                        return_when=concurrent.futures.FIRST_COMPLETED,
                    )
                if (
                    time.time() >= end_time
                    or (errors and fail_fast)
                    or finished
                ):
                    # no time left, exit the loop
                    aborted = True
                    break
                while len(waiting) < num_workers and i < len(values):
                    # space and time available, add tasks
                    value = values[i]
                    future = executor.submit(
                        task, *((value,) + task_args), **task_kwargs,
                    )
                    # small hack to avoid add_done_callback not supporting
                    # extra arguments and closures inside loops retaining
                    # a reference not a value:
                    future._i = i
                    future.add_done_callback(_done_callback)
                    waiting.add(future)
                    i += 1

            if not aborted:
                # all tasks have been submitted, timeout has not been reaches
                # -> wait for all workers to finish before shutting down
                timeout = max(0, end_time - time.time())
                _done, waiting = concurrent.futures.wait(
                    waiting,
                    timeout=timeout,
                    return_when=concurrent.futures.ALL_COMPLETED
                )
            shutdown_executor(executor, waiting)
    finally:
        os.environ['QUTIP_IN_PARALLEL'] = 'FALSE'

    progress_bar.finished()
    if errors and fail_fast:
        raise list(errors.values())[0]
    elif errors:
        raise MapExceptions(
            f"{len(errors)} iterations failed in parallel_map",
            errors, results
        )

    return results


def parallel_map(task, values, task_args=None, task_kwargs=None,
                 reduce_func=None, map_kw=None,
                 progress_bar=None, progress_bar_kwargs={}):
    """
    Parallel execution of a mapping of ``values`` to the function ``task``.
    This is functionally equivalent to::

        result = [task(value, *task_args, **task_kwargs) for value in values]

    Parameters
    ----------
    task : a Python function
        The function that is to be called for each value in ``task_vec``.
    values : array / list
        The list or array of values for which the ``task`` function is to be
        evaluated.
    task_args : list, optional
        The optional additional arguments to the ``task`` function.
    task_kwargs : dictionary, optional
        The optional additional keyword arguments to the ``task`` function.
    reduce_func : func, optional
        If provided, it will be called with the output of each task instead of
        storing them in a list. Note that the order in which results are
        passed to ``reduce_func`` is not defined. It should return None or a
        number. When returning a number, it represents the estimation of the
        number of tasks left. On a return <= 0, the map will end early.
    progress_bar : str, optional
        Progress bar options's string for showing progress.
    progress_bar_kwargs : dict, optional
        Options for the progress bar.
    map_kw: dict, optional
        Dictionary containing entry for:
        - timeout: float, Maximum time (sec) for the whole map.
        - num_cpus: int, Number of jobs to run at once.
        - fail_fast: bool, Abort at the first error.

    Returns
    -------
    result : list
        The result list contains the value of
        ``task(value, *task_args, **task_kwargs)`` for
        each value in ``values``. If a ``reduce_func`` is provided, and empty
        list will be returned.

    """

    map_kw = _read_map_kw(map_kw)
    if sys.version_info >= (3, 7):
        # ProcessPoolExecutor only supports mp_context from 3.7 onwards
        ctx_kw = {"mp_context": mp_context}
    else:
        ctx_kw = {}

    def setup_executor():
        return concurrent.futures.ProcessPoolExecutor(
            max_workers=map_kw['num_cpus'], **ctx_kw,
        )

    def extract_result(future: concurrent.futures.Future):
        exception = future.exception()
        if exception is not None:
            return None, exception
        return future.result(), None

    def shutdown_executor(executor, _):
        # Since `ProcessPoolExecutor` leaves no other option,
        # we wait for all worker processes to finish their current task
        executor.shutdown()

    return _generic_pmap(
        task, values, task_args, task_kwargs, reduce_func,
        map_kw['timeout'], map_kw['fail_fast'], map_kw['num_cpus'],
        progress_bar, progress_bar_kwargs,
        setup_executor, extract_result, shutdown_executor
    )


def loky_pmap(task, values, task_args=None, task_kwargs=None,
              reduce_func=None, map_kw=None,
              progress_bar=None, progress_bar_kwargs={}):
    """
    Parallel execution of a mapping of ``values`` to the function ``task``.
    This is functionally equivalent to::

        result = [task(value, *task_args, **task_kwargs) for value in values]

    Use the loky module instead of multiprocessing.

    Parameters
    ----------
    task : a Python function
        The function that is to be called for each value in ``task_vec``.
    values : array / list
        The list or array of values for which the ``task`` function is to be
        evaluated.
    task_args : list, optional
        The optional additional arguments to the ``task`` function.
    task_kwargs : dictionary, optional
        The optional additional keyword arguments to the ``task`` function.
    reduce_func : func, optional
        If provided, it will be called with the output of each task instead of
        storing them in a list. Note that the order in which results are
        passed to ``reduce_func`` is not defined. It should return None or a
        number. When returning a number, it represents the estimation of the
        number of tasks left. On a return <= 0, the map will end early.
    progress_bar : str, optional
        Progress bar options's string for showing progress.
    progress_bar_kwargs : dict, optional
        Options for the progress bar.
    map_kw: dict, optional
        Dictionary containing entry for:
        - timeout: float, Maximum time (sec) for the whole map.
        - num_cpus: int, Number of jobs to run at once.
        - fail_fast: bool, Abort at the first error.

    Returns
    -------
    result : list
        The result list contains the value of
        ``task(value, *task_args, **task_kwargs)`` for
        each value in ``values``. If a ``reduce_func`` is provided, and empty
        list will be returned.

    """

    from loky import get_reusable_executor
    from loky.process_executor import ShutdownExecutorError
    map_kw = _read_map_kw(map_kw)

    def setup_executor():
        return get_reusable_executor(max_workers=map_kw['num_cpus'])

    def extract_result(future: concurrent.futures.Future):
        exception = future.exception()
        if isinstance(exception, ShutdownExecutorError):
            # Task was aborted due to timeout etc
            return None, None
        if exception is not None:
            return None, exception
        return future.result(), None

    def shutdown_executor(executor, active_tasks):
        # If there are still tasks running, we kill all workers in  order to
        # return immediately. Otherwise, `kill_workers` is set to False so
        # that the worker threads can be reused in subsequent loky_pmap calls.
        kill_workers = len(active_tasks) > 0
        executor.shutdown(kill_workers=kill_workers)

    return _generic_pmap(
        task, values, task_args, task_kwargs, reduce_func,
        map_kw['timeout'], map_kw['fail_fast'], map_kw['num_cpus'],
        progress_bar, progress_bar_kwargs,
        setup_executor, extract_result, shutdown_executor
    )


def mpi_pmap(task, values, task_args=None, task_kwargs=None,
             reduce_func=None, map_kw=None,
             progress_bar=None, progress_bar_kwargs={}):
    """
    Parallel execution of a mapping of ``values`` to the function ``task``.
    This is functionally equivalent to::

        result = [task(value, *task_args, **task_kwargs) for value in values]

    Uses the mpi4py module to execute the tasks asynchronously with MPI
    processes. For more information, consult the documentation of mpi4py and
    the mpi4py.MPIPoolExecutor class.

    Note: in keeping consistent with the API of `parallel_map`, the parameter
    determining the number of requested worker processes is called `num_cpus`.
    The value of `map_kw['num_cpus']` is passed to the MPIPoolExecutor as its
    `max_workers` argument.
    If this parameter is not provided, the environment variable
    `QUTIP_NUM_PROCESSES` is used instead. If this environment variable is not
    set either, QuTiP will use default values that might be unsuitable for MPI
    applications.

    Parameters
    ----------
    task : a Python function
        The function that is to be called for each value in ``task_vec``.
    values : array / list
        The list or array of values for which the ``task`` function is to be
        evaluated.
    task_args : list, optional
        The optional additional arguments to the ``task`` function.
    task_kwargs : dictionary, optional
        The optional additional keyword arguments to the ``task`` function.
    reduce_func : func, optional
        If provided, it will be called with the output of each task instead of
        storing them in a list. Note that the order in which results are
        passed to ``reduce_func`` is not defined. It should return None or a
        number. When returning a number, it represents the estimation of the
        number of tasks left. On a return <= 0, the map will end early.
    progress_bar : str, optional
        Progress bar options's string for showing progress.
    progress_bar_kwargs : dict, optional
        Options for the progress bar.
    map_kw: dict, optional
        Dictionary containing entry for:
        - timeout: float, Maximum time (sec) for the whole map.
        - num_cpus: int, Number of jobs to run at once.
        - fail_fast: bool, Abort at the first error.
        All remaining entries of map_kw will be passed to the
        mpi4py.MPIPoolExecutor constructor.

    Returns
    -------
    result : list
        The result list contains the value of
        ``task(value, *task_args, **task_kwargs)`` for
        each value in ``values``. If a ``reduce_func`` is provided, and empty
        list will be returned.

    """

    from mpi4py.futures import MPIPoolExecutor

    # If the provided num_cpus is None, we use the default value instead.
    # We thus intentionally make it impossible to call
    #   MPIPoolExecutor(max_workers=None, ...)
    # in which case mpi4py would determine a default value. That would be
    # useful, but unfortunately mpi4py provides no public API to access the
    # actual number of workers that is used in that case, which we would need.
    worker_number_provided = (
        ((map_kw is not None) and ('num_cpus' in map_kw))
        or 'QUTIP_NUM_PROCESSES' in os.environ)

    map_kw = _read_map_kw(map_kw)
    timeout = map_kw.pop('timeout')
    num_workers = map_kw.pop('num_cpus')
    fail_fast = map_kw.pop('fail_fast')

    if not worker_number_provided:
        warnings.warn(f'mpi_pmap was called without specifying the number of '
                      f'worker processes, using the default {num_workers}')

    def setup_executor():
        return MPIPoolExecutor(max_workers=num_workers, **map_kw)

    def extract_result(future):
        exception = future.exception()
        if exception is not None:
            return None, exception
        return future.result(), None

    def shutdown_executor(executor, _):
        executor.shutdown()

    return _generic_pmap(
        task, values, task_args, task_kwargs, reduce_func,
        timeout, fail_fast, num_workers,
        progress_bar, progress_bar_kwargs,
        setup_executor, extract_result, shutdown_executor
    )


_maps = {
    "parallel_map": parallel_map,
    "parallel": parallel_map,
    "serial_map": serial_map,
    "serial": serial_map,
    "loky": loky_pmap,
    "mpi": mpi_pmap
}


def _get_map(options):
    map_func = _maps[options['map']]

    if map_func == mpi_pmap:
        map_kw = options['mpi_options']
    else:
        map_kw = {}

    return map_func, map_kw



__all__ = ['partial_transpose']

import numpy as np
import scipy.sparse as sp

from . import (
    Qobj, state_index_number, state_number_index, state_number_enumerate,
)
from .core.dimensions import flatten


def partial_transpose(rho, mask, method='dense'):
    """
    Return the partial transpose of a Qobj instance `rho`,
    where `mask` is an array/list with length that equals
    the number of components of `rho` (that is, the length of
    `rho.dims[0]`), and the values in `mask` indicates whether
    or not the corresponding subsystem is to be transposed.
    The elements in `mask` can be boolean or integers `0` or `1`,
    where `True`/`1` indicates that the corresponding subsystem
    should be tranposed.

    Parameters
    ----------

    rho : :class:`.Qobj`
        A density matrix.

    mask : *list* / *array*
        A mask that selects which subsystems should be transposed.

    method : str {"dense", "sparse"}, default: "dense"
        Choice of method. The "sparse" implementation can be faster for
        large and sparse systems (hundreds of quantum states).

    Returns
    -------

    rho_pr: :class:`.Qobj`
        A density matrix with the selected subsystems transposed.

    """

    mask = [int(i) for i in mask]

    if method == 'sparse':
        return _partial_transpose_sparse(rho, mask)
    else:
        return _partial_transpose_dense(rho, mask)


def _partial_transpose_dense(rho, mask):
    """
    Based on Jonas' implementation using numpy.
    Very fast for dense problems.
    """
    nsys = len(mask)
    pt_dims = np.arange(2 * nsys).reshape(2, nsys).T
    pt_idx = np.concatenate([[pt_dims[n, mask[n]] for n in range(nsys)],
                            [pt_dims[n, 1 - mask[n]] for n in range(nsys)]])

    data = (rho.full()
            .reshape(flatten(rho.dims))
            .transpose(pt_idx)
            .reshape(rho.shape))
    return Qobj(data, dims=rho.dims)


def _partial_transpose_sparse(rho, mask):
    """
    Implement the partial transpose using the CSR sparse matrix.
    """

    data = sp.lil_matrix((rho.shape[0], rho.shape[1]), dtype=complex)
    rho_data = rho.to("CSR").data.as_scipy()

    for m in range(len(rho_data.indptr) - 1):

        n1 = rho_data.indptr[m]
        n2 = rho_data.indptr[m + 1]

        psi_A = state_index_number(rho.dims[0], m)

        for idx, n in enumerate(rho_data.indices[n1:n2]):

            psi_B = state_index_number(rho.dims[1], n)

            m_pt = state_number_index(
                rho.dims[1], np.choose(mask, [psi_A, psi_B]))
            n_pt = state_number_index(
                rho.dims[0], np.choose(mask, [psi_B, psi_A]))

            data[m_pt, n_pt] = rho_data.data[n1 + idx]

    return Qobj(data.tocsr(), dims=rho.dims)


def _partial_transpose_reference(rho, mask):
    """
    This is a reference implementation that explicitly loops over
    all states and performs the transpose. It's slow but easy to
    understand and useful for testing.
    """

    A_pt = np.zeros(rho.shape, dtype=complex)

    for psi_A in state_number_enumerate(rho.dims[0]):
        m = state_number_index(rho.dims[0], psi_A)

        for psi_B in state_number_enumerate(rho.dims[1]):
            n = state_number_index(rho.dims[1], psi_B)

            m_pt = state_number_index(
                rho.dims[1], np.choose(mask, [psi_A, psi_B]))
            n_pt = state_number_index(
                rho.dims[0], np.choose(mask, [psi_B, psi_A]))

            A_pt[m_pt, n_pt] = rho.to("CSR").data.as_scipy()[m, n]

    return Qobj(A_pt, dims=rho.dims)



#cython: language_level=3
#cython: boundscheck=False
#cython: wraparound=False
#cython: initializedcheck=False
#cython: cdvision=True
#cython: c_api_binop_methods=True

from .. import data as _data
from qutip.core.cy.coefficient import coefficient_function_parameters
from qutip.core.data cimport Dense, Data, dense
from qutip.core.data.matmul cimport *
from math import nan as Nan
cdef extern from "<complex>" namespace "std" nogil:
    double complex conj(double complex x)

__all__ = ['_ConstantElement', '_EvoElement',
           '_FuncElement', '_MapElement', '_ProdElement']


cdef class _BaseElement:
    """
    The representation of a single time-dependent term in the list of
    terms used by QobjEvo and solvers to describe operators.

    Conceptually each term is given by ``coeff(t) * qobj(t)`` where
    ``coeff`` is a complex coefficient and ``qobj`` is a :obj:`.Qobj`. Both
    are functions of time. :meth:`~_BaseElement.coeff` returns the
    coefficient at ``t``. :meth:`~_BaseElement.qobj` returns the :obj:`.Qobj`.

    For example, a :obj:`.QobjEvo` instance created by::

      QobjEvo([sigmax(), [sigmay(), 'cos(pi * t)']])

    would contain one :obj:`~_BaseElement` instance for the ``sigmax()`` term,
    and one for the ``[sigmay(), 'cos(pi*t)']`` term.

    :obj:`~_BaseElement` defines the interface to time-dependent terms.
    Sub-classes implement terms defined in different ways.
    For example, :obj:`~_ConstantElement` implements a term that
    consists only of a constant :obj:`.Qobj` (i.e. where there is no dependence
    on ``t``), :obj:`~_EvoElement` implements a term that consists of a
    time-dependet :obj:`.Coefficient` times a constant :obj:`.Qobj`, and
    so on.

    .. note::

      There are three methods to return the factors of the term.

      :meth:`~_BaseElement.coeff` and :meth:`~_BaseElement.qobj` return the
      coefficient and operator factors respectively. They are separate to
      avoid constructing an intermediate Python object when called from
      Cython code (which may take as long as the rest of the call).

      :meth:`~BaseElement.data` is equivalent to ``.qobj(t).data`` and
      provides a convenience method for this common operation.

    .. note::

      All :obj:`~_BaseElement` instances are immutable and methods that would
      modify an object return a new instance instead.
    """
    cpdef Data data(self, t):
        """
        Returns the underlying :obj:`~Data` of the :obj:`.Qobj` component
        of the term at time ``t``.

        Parameters
        ----------
        t : double
          The time, ``t``.

        Returns
        -------
        :obj:`~Data`
          The underlying data of the :obj:`.Qobj` component of the term
          at time ``t``.
        """
        raise NotImplementedError(
          "Sub-classes of _BaseElement should implement .data(t)."
        )

    cpdef object qobj(self, t):
        """
        Returns the :obj:`.Qobj` component of the term at time ``t``.

        Parameters
        ----------
        t : float
          The time, ``t``.

        Returns
        -------
        :obj:`.Qobj`
          The :obj:`.Qobj` component of the term at time ``t``.
        """
        raise NotImplementedError(
          "Sub-classes of _BaseElement should implement .qobj(t)."
        )

    cpdef object coeff(self, t):
        """
        Returns the complex coefficient of the term at time ``t``.

        Parameters
        ----------
        t : float
          The time, ``t``.

        Returns
        -------
        complex
          The complex coefficient of the term at time ``t``.
        """
        raise NotImplementedError(
          "Sub-classes of _BaseElement should implement .coeff(t)."
        )

    cdef Data matmul_data_t(_BaseElement self, t, Data state, Data out=None):
        """
        Possibly in-place multiplication and addition. Multiplies a given state
        by the elemen's value at time ``t`` and adds the result to ``out``.
        Equivalent to::

          out += self.coeff(t) * self.qobj(t) @ state

        Sub-classes may override :meth:`~matmul_data_t` to provide an more
        efficient implementation.

        Parameters
        ----------
        t : double
          The time, ``t``.

        state : :obj:`~Data`
          The state to multiply by the element term.

        out : :obj:`~Data` or ``NULL``
          The output to add the result of the multiplication to. If ``NULL``,
          the result of the multiplication is returned directly (i.e. ``out``
          is assumed to be the zero matrix).

        Returns
        -------
        data
          The result of ``self.coeff(t) * self.qobj(t) @ state + out``, with
          the addition possibly having been performed in-place on ``out``.

        .. note::

          Because of the possibly but not definitely in-place behaviour of
          this function, the result should always be assigned to some variable
          and surrounding code should assume that ``out`` maybe have been
          modified and that the result may be a reference to ``out``. The
          safest is simply to write ``out = elem.matmul_data_t(t, state, out)``.

        .. note::

          This method would ideally be implemented using a special
          function dispatched by the data layer so that it did not have to
          special case the ``Dense`` operation itself, but the data layer
          dispatch does not yet support in-place operations. Once it does
          this method should be updated to use the new support.
        """
        if out is None:
            return _data.matmul(self.data(t), state, self.coeff(t))
        elif type(state) is Dense and type(out) is Dense:
            imatmul_data_dense(self.data(t), state, self.coeff(t), out)
            return out
        else:
            return _data.add(
                out,
                _data.matmul(self.data(t), state, self.coeff(t))
            )

    def linear_map(self, f, anti=False):
        """
        Return a new element representing a linear transformation ``f``
        of the :obj:`.Qobj` portion of this element and possibly a
        complex conjucation of the coefficient portion (when ``f`` is an
        antilinear map).

        If this element represents ``coeff * qobj`` the returned element
        represents ``coeff * f(obj)`` (if ``anti=False``) or
        ``conj(coeff) * f(obj)`` (if ``anti=True``).

        Parameters
        ----------
        f : function
          The linear transformation to apply to the :obj:`.Qobj` of this
          element.
        anti : bool
          Whether to take the complex conjugate of the coefficient. Default
          is ``False``. Should be set to ``True`` if ``f`` is an antilinear
          map such as the adjoint (i.e. dagger).

        Returns
        -------
        _BaseElement
          A new element with the transformation applied.
        """
        raise NotImplementedError(
          "Sub-classes of _BaseElement should implement .linear_map(t)."
        )

    def replace_arguments(self, args, cache=None):
        """
        Return a copy of the element with the (possible) additional arguments
        to any time-dependent functions updated to the given argument values.
        The arguments of any contained :obj:`.Coefficient` instances are also
        replaced.

        If the operation does not modify this element, the original element
        may be returned.

        Parameters
        ----------
        args : dict
          A dictionary of arguments to update. Keys are the names of the
          arguments and values are the new argument values. Arguments
          not included retain their original values.

        cache : list or ``None``
          A cache to add updated elements to. Unmodified elements are not
          added to the cache. If a cache is supplied and ``.replace_arguments``
          would be called again on the same element with the same arguments,
          the new element from the cache will be returned instead.

          By default the cache is ``None`` and no caching is performed. Cache
          users should supply either ``cache=[]`` (which activates caching)
          or an existing cache (for example, if making calls on sub-elements
          of some composite element).

        Returns
        -------
        _BaseElement
          A new element with the arguments replaced, or possibly this element
          if it would not be modified.

        Example
        -------

        If ``elem`` is a product element for ``op * op.dag()`` it will contain
        two references to the same element for ``op``. Calling::

          elem = elem.replace_arguments({"theta": 3.1416})

        will return a new element that contains two *different* copies of
        ``op.replace_arguments({"theta": 3.1416})``. This will cause ``op``
        to be evaluated *twice* when ``elem.qobj(t)`` is called.

        Calling instead::

          elem = elem.replace_arguments({"theta": 3.1416}, cache=[])

        will return a new element that contains two references to *one* copy
        of ``op.replace_arguments({"theta": 3.1416})``, which will improve
        performance when calling ``elem.qobj(t)`` later.
        """
        raise NotImplementedError(
          "Sub-classes of _BaseElement should implement .replace_arguments(t)."
        )

    def __call__(self, t, args=None):
        if args:
            cache = []
            new = self.replace_arguments(args, cache)
            return new.qobj(t) * new.coeff(t)
        return self.qobj(t) * self.coeff(t)

    @property
    def dtype(self):
        return None


cdef class _ConstantElement(_BaseElement):
    """
    Constant part of a list format :obj:`.QobjEvo`.
    A constant :obj:`.QobjEvo` will contain one `_ConstantElement`::

      qevo = QobjEvo(H0)
      qevo.elements = [_ConstantElement(H0)]
    """
    def __init__(self, qobj):
        self._qobj = qobj
        self._data = self._qobj.data

    def __mul__(left, right):
        if type(left) is _ConstantElement:
            return _ConstantElement((<_ConstantElement> left)._qobj * right)
        elif type(right) is _ConstantElement:
            return _ConstantElement((<_ConstantElement> right)._qobj * left)
        return NotImplemented

    def __matmul__(left, right):
        if type(left) is _ConstantElement and type(right) is _ConstantElement:
            return _ConstantElement(
                (<_ConstantElement> left)._qobj @
                (<_ConstantElement> right)._qobj
            )
        return NotImplemented

    cpdef Data data(self, t):
        return self._data

    cpdef object qobj(self, t):
        return self._qobj

    cpdef object coeff(self, t):
        return 1.

    def linear_map(self, f, anti=False):
        return _ConstantElement(f(self._qobj))

    def replace_arguments(self, args, cache=None):
        return self

    def __call__(self, t, args=None):
        return self._qobj

    @property
    def dtype(self):
        return type(self._data)


cdef class _EvoElement(_BaseElement):
    """
    A pair of a :obj:`.Qobj` and a :obj:`.Coefficient` from the list format
    time-dependent operator::

      qevo = QobjEvo([[H0, coeff0], [H1, coeff1]])
      qevo.elements = [_EvoElement(H0, coeff0), _EvoElement(H1, coeff1)]
    """
    def __init__(self, qobj, coefficient):
        self._qobj = qobj
        self._data = self._qobj.data
        self._coefficient = coefficient

    def __mul__(left, right):
        cdef _EvoElement base
        cdef object factor
        if type(left) is _EvoElement:
            base = left
            factor = right
        if type(right) is _EvoElement:
            base = right
            factor = left
        return _EvoElement(base._qobj * factor, base._coefficient)

    def __matmul__(left, right):
        if isinstance(left, _EvoElement) and isinstance(right, _EvoElement):
            coefficient = left._coefficient * right._coefficient
        elif isinstance(left, _EvoElement) and isinstance(right, _ConstantElement):
            coefficient = left._coefficient
        elif isinstance(right, _EvoElement) and isinstance(left, _ConstantElement):
            coefficient = right._coefficient
        else:
            return NotImplemented
        return _EvoElement(left._qobj * right._qobj, coefficient)

    cpdef Data data(self, t):
        return self._data

    cpdef object qobj(self, t):
        return self._qobj

    cpdef object coeff(self, t):
        return self._coefficient(t)

    def linear_map(self, f, anti=False):
        return _EvoElement(
            f(self._qobj),
            self._coefficient.conj() if anti else self._coefficient,
        )

    def replace_arguments(self, args, cache=None):
        return _EvoElement(
            self._qobj,
            self._coefficient.replace_arguments(args)
        )

    @property
    def dtype(self):
        return type(self._data)


cdef class _FuncElement(_BaseElement):
    """
    Used with :obj:`.QobjEvo` to build an evolution term from a function with
    either the signature: ::

        func(t: float, ...) -> Qobj

    or the older QuTiP 4 style signature: ::

        func(t: float, args: dict) -> Qobj

    In the new style, ``func`` may accept arbitrary named arguments and
    is called as ``func(t, **args)``.

    A :obj:`.QobjEvo` created from such a function contains one
    :obj:`_FuncElement`: ::

        qevo = QobjEvo(func, args=args)
        qevo.elements = [_FuncElement(func, args)]

    Each :obj:`_FuncElement` consists of an immutable pair ``(func, args)``
    of function and argument.

    This class has a basic capacity to memoize calls to ``func`` by saving the
    result of the last call ::

        op = QobjEvo(func, args=args)
        (op.dag() * op)(t)

    calls ``func`` only once.

    Parameters
    ----------
    func : callable(t : float, ...) -> Qobj
        Function returning the element value at time ``t``.

    args : dict
        Values of the arguments to pass to ``func``.

    style : {None, "pythonic", "dict", "auto"}
        The style of the signature used. If style is ``None``,
        the value of ``qutip.settings.core["function_coefficient_style"]``
        is used. Otherwise the supplied value overrides the global setting.

    The parameters ``_f_pythonic`` and ``_f_parameters`` override function
    style and parameter detection and are not intended to be part of
    the public interface.
    """

    _UNSET = object()

    def __init__(self, func, args, style=None, _f_pythonic=_UNSET, _f_parameters=_UNSET):
        if _f_pythonic is self._UNSET or _f_parameters is self._UNSET:
            if not (_f_pythonic is self._UNSET and _f_parameters is self._UNSET):
                raise TypeError(
                    "_f_pythonic and _f_parameters should always be given together."
                )
            _f_pythonic, _f_parameters = coefficient_function_parameters(
                func, style=style)
            if _f_parameters is not None:
                args = {k: args[k] for k in _f_parameters & args.keys()}
            else:
                args = args.copy()
        self._func = func
        self._args = args
        self._f_pythonic = _f_pythonic
        self._f_parameters = _f_parameters
        self._previous = (Nan, None)

    def __mul__(left, right):
        cdef _MapElement out
        if type(left) is _FuncElement:
            out = _MapElement(left, [], right)
        if type(right) is _FuncElement:
            out = _MapElement(right, [], left)
        return out

    def __matmul__(left, right):
        return _ProdElement(left, right, [])

    cpdef Data data(self, t):
        return self.qobj(t).data

    cpdef object qobj(self, t):
        cdef double _t
        cdef object _qobj
        _t, _qobj = self._previous
        if t == _t:
            return _qobj
        if self._f_pythonic:
            _qobj = self._func(t, **self._args)
        else:
            _qobj = self._func(t, self._args)
        self._previous = (t, _qobj)
        return _qobj

    cpdef object coeff(self, t):
        return 1.

    def linear_map(self, f, anti=False):
        return _MapElement(self, [f])

    def replace_arguments(_FuncElement self, args, cache=None):
        if self._f_parameters is not None:
            args = {k: args[k] for k in self._f_parameters & args.keys()}
        if not args:
            return self
        if cache is not None:
            for old, new in cache:
                if old is self:
                    return new
        new = _FuncElement(
                self._func,
                {**self._args, **args},
                _f_pythonic=self._f_pythonic,
                _f_parameters=self._f_parameters,
        )
        if cache is not None:
            cache.append((self, new))
        return new


cdef class _MapElement(_BaseElement):
    """
    :obj:`_FuncElement` decorated with linear tranformations.

    Linear tranformations available in :obj:`.QobjEvo` include transpose,
    adjoint, conjugate, convertion and product with number::
    ```
        op = QobjEvo(f, args=args)
        op2 = op.conj().dag() * 2
    ```
    Then ``op2.elements`` is::
        ``[_MapElement(_FuncElement(f, args), [conj, dag], 2)]``

    """
    def __init__(self, _FuncElement base, transform, coeff=1.):
        self._base = base
        self._transform = transform
        self._coeff = coeff

    def __mul__(left, right):
        cdef _MapElement out, self
        cdef double complex factor
        if type(left) is _MapElement:
            self = left
            factor = right
        elif type(right) is _MapElement:
            self = right
            factor = left
        return _MapElement(
            self._base,
            self._transform.copy(),
            self._coeff*factor
        )

    def __matmul__(left, right):
        return _ProdElement(left, right, [])

    cpdef Data data(self, t):
        return self.qobj(t).data

    cpdef object qobj(self, t):
        out = self._base.qobj(t)
        for func in self._transform:
            out = func(out)
        return out

    cpdef object coeff(self, t):
        return self._coeff

    def linear_map(self, f, anti=False):
        return _MapElement(
            self._base,
            self._transform + [f],
            conj(self._coeff) if anti else self._coeff
        )

    def replace_arguments(_MapElement self, args, cache=None):
        return _MapElement(
            self._base.replace_arguments(args, cache=cache),
            self._transform.copy(),
            self._coeff,
        )


cdef class _ProdElement(_BaseElement):
    """
    Product of a :obj:`_FuncElement` or :obj:`_MapElement` with other
    :obj:`_BaseElement`. Include a stack of linear transformation to be
    applied after the product::

        ``op = QobjEvo(f) * qobj1``
    Then ``op.elements`` is::
        ``[_ProdElement(_FuncElement(f, {}), _ConstantElement(qobj1))]``
    """
    def __init__(self, left, right, transform, conj=False):
        self._left = left
        self._right = right
        self._conj = conj
        self._transform = transform

    def __mul__(left, right):
        cdef _ProdElement self
        cdef double complex factor
        if type(left) is _ProdElement:
            self = left
            factor = right
        if type(right) is _ProdElement:
            self = right
            factor = left
        return _ProdElement(self._left, self._right * factor,
                            self._transform.copy(), self._conj)

    def __matmul__(left, right):
        return _ProdElement(left, right, [])

    cpdef Data data(self, t):
        return self.qobj(t).data

    cpdef object qobj(self, t):
        out = self._left.qobj(t) @ self._right.qobj(t)
        for func in self._transform:
            out = func(out)
        return out

    cpdef object coeff(self, t):
        cdef double complex out = self._left.coeff(t) * self._right.coeff(t)
        return conj(out) if self._conj else out

    cdef Data matmul_data_t(_ProdElement self, t, Data state, Data out=None):
        cdef Data temp
        if not self._transform:
            temp = self._right.matmul_data_t(t, state)
            out = self._left.matmul_data_t(t, temp, out)
            return out
        elif type(state) is Dense and type(out) is Dense:
            imatmul_data_dense(self.data(t), state, self.coeff(t), out)
            return out
        else:
            return _data.add(
                out,
                _data.matmul(self.data(t), state, self.coeff(t))
            )

    def linear_map(self, f, anti=False):
        return _ProdElement(
            self._left, self._right,
            self._transform + [f],
            self._conj ^ anti
        )

    def replace_arguments(_ProdElement self, args, cache=None):
        return _ProdElement(
            self._left.replace_arguments(args, cache=cache),
            self._right.replace_arguments(args, cache=cache),
            self._transform.copy(),
            self._conj
        )



""" Floquet solver compatibility functions that behave like the corresponding
functions from QuTiP 4.7.

These functions are indented to be used when porting code from QuTiP 4.7 to
QuTiP 5. They are deprecated and will be removed in QuTiP 5.1.
"""

__all__ = [
    "floquet_modes",
    "floquet_modes_t",
    "floquet_modes_table",
    "floquet_modes_t_lookup",
    "floquet_states",
    "floquet_states_t",
    "floquet_wavefunction",
    "floquet_wavefunction_t",
    "floquet_state_decomposition",
    "floquet_master_equation_rates",
]

from .floquet import *
import numpy as np
import warnings


def floquet_modes(H, T, args=None, sort=False, U=None, options=None):
    """
    Calculate the initial Floquet modes Phi_alpha(0) for a driven system with
    period T.

    Deprecated from qutip v5. Use :class:`.FloquetBasis` instead:

    fbasis = FloquetBasis(H, T, args=args, options=options, sort=sort)
    f_mode_0 = fbasis.mode(0)
    f_energies = fbasis.e_quasi
    """
    warnings.warn(FutureWarning(
        "`floquet_modes` is deprecated. Use `FloquetBasis.mode` instead."
    ))
    fbasis = FloquetBasis(H, T, args=args, options=options, sort=sort)
    f_mode_0 = fbasis.mode(0)
    f_energies = fbasis.e_quasi
    return f_mode_0, f_energies


def floquet_modes_t(f_modes_0, f_energies, t, H, T, args=None, options=None):
    """
    Calculate the Floquet modes at times tlist Phi_alpha(tlist) propagting the
    initial Floquet modes Phi_alpha(0).

    Deprecated from qutip v5. Use :class:`.FloquetBasis` instead:

    fbasis = FloquetBasis(H, T, args=args, options=options)
    f_mode_t = fbasis.mode(t)
    """
    warnings.warn(FutureWarning(
        "`floquet_modes_t` is deprecated. Use `FloquetBasis.mode` instead."
    ))
    fbasis = FloquetBasis(H, T, args=args, options=options)
    return fbasis.mode(t)


def floquet_modes_table(
    f_modes_0, f_energies, tlist, H, T, args=None, options=None
):
    """
    Pre-calculate the Floquet modes for a range of times spanning the floquet
    period. Can later be used as a table to look up the floquet modes for
    any time.

    Deprecated from qutip v5. Use :class:`.FloquetBasis` instead:

    fbasis = FloquetBasis(H, T, args=args, options=options, precompute=tlist)
    """
    raise NotImplementedError("`floquet_modes_table` is deprecated.")


def floquet_modes_t_lookup(f_modes_table_t, t, T):
    """
    Lookup the floquet mode at time t in the pre-calculated table of floquet
    modes in the first period of the time-dependence.

    Deprecated from qutip v5. Use :class:`.FloquetBasis` instead:

    f_modes_table_t = fbasis = FloquetBasis(...)
    f_mode_t = f_modes_table_t.mode(t)
    """
    raise NotImplementedError(
        "`floquet_modes_t_lookup` is no longer provided. "
        "Use `FloquetBasis` instead."
    )


def floquet_states(f_modes_t, f_energies, t):
    """
    Evaluate the floquet states at time t given the Floquet modes at that time.

    Deprecated from qutip v5. Use :class:`.FloquetBasis` instead:

    fbasis = FloquetBasis(H, T, args=args, options=options)
    f_state_t = fbasis.state(t)
    """
    warnings.warn(FutureWarning(
        "`floquet_states` is deprecated. "
        "Use `FloquetBasis.state` instead."
    ))
    return [
        (f_modes_t[i] * np.exp(-1j * f_energies[i] * t))
        for i in np.arange(len(f_energies))
    ]


def floquet_states_t(f_modes_0, f_energies, t, H, T, args=None, options=None):
    """
    Evaluate the floquet states at time t given the initial Floquet modes.

    Deprecated from qutip v5. Use :class:`.FloquetBasis` instead:

    fbasis = FloquetBasis(H, T, args=args, options=options)
    f_state_t = fbasis.state(t)
    """
    warnings.warn(FutureWarning(
        "`floquet_states_t` is deprecated. "
        "Use `FloquetBasis.state` instead."
    ))
    fbasis = FloquetBasis(H, T, args=args, options=options)
    return fbasis.state(t)


def floquet_wavefunction(f_modes_t, f_energies, f_coeff, t):
    """
    Evaluate the wavefunction for a time t using the Floquet state
    decompositon, given the Floquet modes at time `t`.

    Deprecated from qutip v5. Use :class:`.FloquetBasis` instead:

    fbasis = FloquetBasis(H, T, args=args, options=options)
    psi_t = fbasis.from_floquet_basis(f_coeff, t)
    """
    raise NotImplementedError(
        "`floquet_wavefunction` is not longer provided. "
        "Use `FloquetBasis.from_floquet_basis` instead."
    )


def floquet_wavefunction_t(
    f_modes_0, f_energies, f_coeff, t, H, T, args=None, options=None
):
    """
    Evaluate the wavefunction for a time t using the Floquet state
    decompositon, given the initial Floquet modes.

    Deprecated from qutip v5. Use :class:`.FloquetBasis` instead:

    fbasis = FloquetBasis(H, T, args=args, options=options)
    psi_t = fbasis.from_floquet_basis(f_coeff, t)
    """
    warnings.warn(FutureWarning(
        "`floquet_wavefunction_t` is deprecated. "
        "Use `FloquetBasis.from_floquet_basis` instead."
    ))
    fbasis = FloquetBasis(H, T, args=args, options=options)
    return fbasis.from_floquet_basis(f_coeff, t)


def floquet_state_decomposition(f_states, f_energies, psi):
    r"""
    Decompose the wavefunction `psi` (typically an initial state) in terms of
    the Floquet states, :math:`\psi = \sum_\alpha c_\alpha \psi_\alpha(0)`.

    Deprecated from qutip v5. Use :class:`.FloquetBasis` instead:

    fbasis = FloquetBasis(H, T, args=args, options=options)
    f_coeff = fbasis.to_floquet_basis(psi)
    """
    raise NotImplementedError(
        "`floquet_state_decomposition` is deprecated. "
        "Use `FloquetBasis.to_floquet_basis` instead."
    )


def floquet_master_equation_rates(
    f_modes_0,
    f_energies,
    c_op,
    H,
    T,
    args,
    J_cb,
    w_th,
    kmax=5,
    f_modes_table_t=None,
):
    """
    Calculate the rates and matrix elements for the Floquet-Markov master
    equation.

    .. note ::

        Deprecated. For the Floquet-Markov master equation's tensor, use
        :func:`floquet_tensor`. For the rates matrices, use
        :func:`floquet_delta_tensor`, :func:`floquet_X_matrices`,
        :func:`floquet_gamma_matrices` and/or s:func:`floquet_A_matrix`.

    Parameters
    ----------
    f_modes_0 : Any
        No longer used.
    f_energies : Any
        No longer used.
    c_op : :class:`.Qobj`
        The collapse operators describing the dissipation.
    H : :class:`.Qobj`
        System Hamiltonian, time-dependent with period `T`.
    T : float
        The period of the time-dependence of the hamiltonian.
    args : dictionary
        Dictionary with variables required to evaluate H.
    J_cb : callback functions
        A callback function that computes the noise power spectrum, as
        a function of frequency, associated with the collapse operator `c_op`.
    w_th : float
        The temperature in units of frequency.
    kmax : int, default=5
        The truncation of the number of sidebands.
    f_modes_table_t : Any
        No longer used.

    Returns
    -------
    output : list
        A list (Delta, X, Gamma, A) containing the matrices Delta, X, Gamma
        and A used in the construction of the Floquet-Markov master equation.
    """
    warnings.warn(
        FutureWarning("`floquet_master_equation_rates` is deprecated.")
    )
    floquet_basis = FloquetBasis(H, T, args=args)
    energy = floquet_basis.e_quasi
    delta = floquet_delta_tensor(energy, kmax, T)
    x = floquet_X_matrices(floquet_basis, [c_op], kmax, nT)
    gamma = floquet_gamma_matrices(x, delta, [J_cb])
    a = floquet_A_matrix(delta, gamma, w_th)
    return delta, x[0], gamma, a



# -*- coding: utf-8 -*-
"""
Created on Wed May 29 11:23:46 2013

@author: dcriger
"""

from __future__ import division

import numpy as np
import pytest

from qutip import (
    Qobj, basis, identity, sigmax, sigmay, qeye, create, rand_super,
    rand_super_bcsz, rand_dm, tensor, super_tensor, kraus_to_choi,
    to_super, to_choi, to_kraus, to_chi, to_stinespring, operator_to_vector,
    vector_to_operator, sprepost, destroy, CoreOptions
)
from qutip.core.gates import swap


tol = 1e-8


def assert_kraus_equivalence(a, b, tol=tol):
    assert a.shape == b.shape
    assert a.dims == b.dims
    assert a.type == b.type
    # Kraus operators may vary by a global phase.  Let's find the first
    # non-zero element, set the phase such that that element has the same phase
    # in both, and then compare.  We have to take care to find the first
    # element.
    a, b = a.full(), b.full()
    a_nz = np.nonzero(np.abs(a) > tol)
    if len(a_nz[0]) == 0:
        np.testing.assert_allclose(b, 0, atol=tol)
    a_el, b_el = a[a_nz[0][0], a_nz[1][0]], b[a_nz[0][0], a_nz[1][0]]
    assert b_el != 0
    b *= (a_el / np.abs(a_el)) / (b_el / np.abs(b_el))
    np.testing.assert_allclose(a, b)


@pytest.fixture(scope="function", params=[2, 3, 7])
def dimension(request):
    # There are also some cases in the file where this fixture is explicitly
    # overridden by a more local mark.  That is deliberate.
    return request.param


@pytest.fixture(scope="function", params=[
    pytest.param(rand_super, id="super"),
    pytest.param(rand_super_bcsz, id="super_bcz")
])
def superoperator(request, dimension):
    return request.param(dimension)


class TestSuperopReps:
    """
    A test class for the QuTiP function for applying superoperators to
    subsystems.
    """

    def test_SuperChoiSuper(self, superoperator):
        """
        Superoperator: Converting superoperator to Choi matrix and back.
        """
        choi_matrix = to_choi(superoperator)
        test_supe = to_super(choi_matrix)

        # Assert both that the result is close to expected, and has the right
        # type.
        assert (test_supe - superoperator).norm() < tol
        assert choi_matrix.type == "super" and choi_matrix.superrep == "choi"
        assert test_supe.type == "super" and test_supe.superrep == "super"

    @pytest.mark.parametrize('dimension', [2, 4])
    def test_SuperChoiChiSuper(self, dimension):
        """
        Superoperator: Converting two-qubit superoperator through
        Choi and chi representations goes back to right superoperator.
        """
        superoperator = super_tensor(
            rand_super(dimension), rand_super(dimension),
        )

        choi_matrix = to_choi(superoperator)
        chi_matrix = to_chi(choi_matrix)
        test_supe = to_super(chi_matrix)

        # Assert both that the result is close to expected, and has the right
        # type.
        assert (test_supe - superoperator).norm() < tol
        assert choi_matrix.type == "super" and choi_matrix.superrep == "choi"
        assert chi_matrix.type == "super" and chi_matrix.superrep == "chi"
        assert test_supe.type == "super" and test_supe.superrep == "super"

    def test_ChoiKrausChoi(self, superoperator):
        """
        Superoperator: Convert superoperator to Choi matrix and back.
        """
        choi_matrix = to_choi(superoperator)
        kraus_ops = to_kraus(choi_matrix)
        test_choi = kraus_to_choi(kraus_ops)

        # Assert both that the result is close to expected, and has the right
        # type.
        assert (test_choi - choi_matrix).norm() < tol
        assert choi_matrix.type == "super" and choi_matrix.superrep == "choi"
        assert test_choi.type == "super" and test_choi.superrep == "choi"

    def test_NonSquareKrausSuperChoi(self):
        """
        Superoperator: Convert non-square Kraus operator to Super + Choi matrix
        and back.
        """
        zero = np.array([[1], [0]], dtype=complex)
        one = np.array([[0], [1]], dtype=complex)
        zero_log = np.kron(np.kron(zero, zero), zero)
        one_log = np.kron(np.kron(one, one), one)
        # non-square Kraus operator (isometry)
        kraus = Qobj(zero_log @ zero.T + one_log @ one.T)
        super = sprepost(kraus, kraus.dag())
        choi = to_choi(super)
        op1 = to_kraus(super)
        op2 = to_kraus(choi)
        op3 = to_super(choi)
        assert choi.type == "super" and choi.superrep == "choi"
        assert super.type == "super" and super.superrep == "super"
        assert_kraus_equivalence(op1[0], kraus, tol=1e-8)
        assert_kraus_equivalence(op2[0], kraus, tol=1e-8)
        assert abs((op3 - super).norm()) < 1e-8

    def test_NeglectSmallKraus(self):
        """
        Superoperator: Convert Kraus to Choi matrix and back. Neglect tiny
        Kraus operators.
        """
        zero = np.array([[1], [0]], dtype=complex)
        one = np.array([[0], [1]], dtype=complex)
        zero_log = np.kron(np.kron(zero, zero), zero)
        one_log = np.kron(np.kron(one, one), one)
        # non-square Kraus operator (isometry)
        kraus = Qobj(zero_log @ zero.T + one_log @ one.T)
        super = sprepost(kraus, kraus.dag())
        # 1 non-zero Kraus operator the rest are zero
        sixteen_kraus_ops = to_kraus(super, tol=0.0)
        # default is tol=1e-9
        one_kraus_op = to_kraus(super)
        assert len(sixteen_kraus_ops) == 16
        assert len(one_kraus_op) == 1
        assert_kraus_equivalence(one_kraus_op[0], kraus, tol=tol)

    def test_SuperPreservesSelf(self, superoperator):
        """
        Superoperator: to_super(q) returns q if q is already a
        supermatrix.
        """

        assert superoperator is to_super(superoperator)

    def test_ChoiPreservesSelf(self, superoperator):
        """
        Superoperator: to_choi(q) returns q if q is already Choi.
        """
        choi = to_choi(superoperator)
        assert choi is to_choi(choi)

    def test_random_iscptp(self, superoperator):
        """
        Superoperator: Randomly generated superoperators are
        correctly reported as CPTP and HP.
        """
        with CoreOptions(atol=1e-9):
            assert superoperator.iscptp
            assert superoperator.ishp

    @pytest.mark.parametrize(['qobj', 'hp', 'cp', 'tp'], [
        pytest.param(sprepost(destroy(2), create(2)), True, True, False),
        pytest.param(sprepost(destroy(2), destroy(2)), False, False, False),
        pytest.param(qeye(2), True, True, True),
        pytest.param(sigmax(), True, True, True),
        pytest.param(tensor(sigmax(), qeye(2)), True, True, True),
        pytest.param(0.5 * (to_super(tensor(sigmax(), qeye(2)))
                            + to_super(tensor(qeye(2), sigmay()))),
                     True, True, True,
                     id="linear combination of bipartite unitaries"),
        pytest.param(Qobj(swap(), dims=[[[2],[2]]]*2, superrep='choi'),
                     True, False, True,
                     id="partial transpose map"),
        pytest.param(Qobj(qeye(4)*0.9, dims=[[[2],[2]]]*2), True, True, False,
                     id="subnormalized map"),
        pytest.param(basis(2, 0), False, False, False, id="ket"),
    ])
    def test_known_iscptp(self, qobj, hp, cp, tp):
        """
        Superoperator: ishp, iscp, istp and iscptp known cases.
        """
        assert qobj.ishp == hp
        assert qobj.iscp == cp
        assert qobj.istp == tp
        assert qobj.iscptp == (cp and tp)

    def test_choi_tr(self):
        """
        Superoperator: Trace returned by to_choi matches docstring.
        """
        for dims in range(2, 5):
            assert abs(to_choi(identity(dims)).tr() - dims) < tol


    # Conjugation by a creation operator
    a = create(2).dag()
    S = sprepost(a, a.dag())

    # A single off-diagonal element
    S_ = sprepost(a, a)

    # Check that a linear combination of bipartite unitaries is CPTP and HP.
    S_U = (
        to_super(tensor(sigmax(), identity(2))) +
        to_super(tensor(identity(2), sigmay()))
    ) / 2

    # The partial transpose map, whose Choi matrix is SWAP
    ptr_swap = Qobj(swap(), dims=[[[2], [2]]]*2, superrep='choi')

    # Subnormalized maps (representing erasure channels, for instance)
    subnorm_map = Qobj(identity(4) * 0.9, dims=[[[2], [2]]]*2,
                       superrep='super')

    @pytest.mark.parametrize(['qobj',  'shouldhp', 'shouldcp', 'shouldtp'], [
        pytest.param(S, True, True, False, id="conjugatio by create op"),
        pytest.param(S_, False, False, False, id="single off-diag"),
        pytest.param(identity(2), True, True, True, id="Identity"),
        pytest.param(sigmax(), True, True, True, id="Pauli X"),
        pytest.param(
            tensor(sigmax(), identity(2)), True, True, True,
            id="bipartite system",
        ),
        pytest.param(
            S_U, True, True, True, id="linear combination of bip. unitaries",
        ),
        pytest.param(ptr_swap,  True, False, True, id="partial transpose map"),
        pytest.param(subnorm_map, True, True, False, id="subnorm map"),
        pytest.param(basis(2), False, False, False, id="not an operator"),

    ])
    def test_known_iscptp(self, qobj, shouldhp, shouldcp, shouldtp):
        """
        Superoperator: ishp, iscp, istp and iscptp known cases.
        """
        assert qobj.ishp == shouldhp
        assert qobj.iscp == shouldcp
        assert qobj.istp == shouldtp
        assert qobj.iscptp == (shouldcp and shouldtp)

    def test_choi_tr(self, dimension):
        """
        Superoperator: Trace returned by to_choi matches docstring.
        """
        assert abs(to_choi(identity(dimension)).tr() - dimension) <= tol

    def test_stinespring_cp(self, dimension):
        """
        Stinespring: A and B match for CP maps.
        """
        superop = rand_super_bcsz(dimension)
        A, B = to_stinespring(superop)

        assert (A - B).norm() < tol

    @pytest.mark.repeat(3)
    def test_stinespring_agrees(self, dimension):
        """
        Stinespring: Partial Tr over pair agrees w/ supermatrix.
        """

        map = rand_super_bcsz(dimension)
        state = rand_dm(dimension)

        S = to_super(map)
        A, B = to_stinespring(map)

        q1 = vector_to_operator(
            S * operator_to_vector(state)
        )
        # FIXME: problem if Kraus index is implicitly
        #        ptraced!
        q2 = (A * state * B.dag()).ptrace((0,))

        assert (q1 - q2).norm('tr') <= tol

    def test_stinespring_dims(self, dimension):
        """
        Stinespring: Check that dims of channels are preserved.
        """
        chan = super_tensor(to_super(sigmax()), to_super(qeye(dimension)))
        A, B = to_stinespring(chan)
        assert A.dims == [[2, dimension, 1], [2, dimension]]
        assert B.dims == [[2, dimension, 1], [2, dimension]]

    @pytest.mark.parametrize('dimension', [2, 4, 8])
    def test_chi_choi_roundtrip(self, dimension):

        superop = rand_super_bcsz(dimension)
        superop = to_chi(superop)
        rt_superop = to_chi(to_choi(superop))
        dif = (rt_superop - superop).norm()

        assert dif == pytest.approx(0, abs=1e-7)
        assert rt_superop.type == superop.type
        assert rt_superop.dims == superop.dims

    chi_sigmax = [
        [0, 0, 0, 0],
        [0, 4, 0, 0],
        [0, 0, 0, 0],
        [0, 0, 0, 0]
    ]
    chi_diag2 = [
        [4, 0, 0, 0],
        [0, 0, 0, 0],
        [0, 0, 0, 0],
        [0, 0, 0, 0]
    ]
    rotX_pi_4 = (-1j * sigmax() * np.pi / 4).expm()
    chi_rotX_pi_4 = [
        [2, 2j, 0, 0],
        [-2j, 2, 0, 0],
        [0, 0, 0, 0],
        [0, 0, 0, 0]
    ]

    @pytest.mark.parametrize(['superop', 'chi_expected'], [
        pytest.param(sigmax(), chi_sigmax),
        pytest.param(to_super(sigmax()), chi_sigmax),
        pytest.param(qeye(2), chi_diag2),
        pytest.param(rotX_pi_4, chi_rotX_pi_4)
    ])
    def test_chi_known(self, superop, chi_expected):
        """
        Superoperator: Chi-matrix for known cases is correct.
        """
        chi_actual = to_chi(superop)
        chiq = Qobj(
            chi_expected, dims=[[[2], [2]], [[2], [2]]], superrep='chi',
        )
        assert (chi_actual - chiq).norm() < tol



import os
import warnings

import qutip.settings
from qutip.settings import settings
import qutip.version
from qutip.version import version as __version__
# -----------------------------------------------------------------------------
# Look to see if we are running with OPENMP
#
# Set environ variable to determin if running in parallel mode
# (i.e. in parfor or parallel_map)
os.environ['QUTIP_IN_PARALLEL'] = 'FALSE'


# -----------------------------------------------------------------------------
# Check that import modules are compatible with requested configuration
#

# Check for Matplotlib
try:
    import matplotlib
except ImportError:
    warnings.warn("matplotlib not found: Graphics will not work.")
else:
    del matplotlib


# -----------------------------------------------------------------------------
# Load modules
#

from .core import *
from .solver import *
from .solver import nonmarkov
import qutip.piqs.piqs as piqs

# graphics
from .bloch import *
from .visualization import *
from .animation import *
from .matplotlib_utilities import *

# library functions
from .tomography import *
from .wigner import *
from .random_objects import *
from .simdiag import *
from .entropy import *
from .partial_transpose import *
from .continuous_variables import *
from .distributions import *
from . import measurement

# utilities
from .utilities import *
from .fileio import *
from .about import *
from .cite import *

# -----------------------------------------------------------------------------
# Clean name space
#
del os, warnings



import pytest
import pickle
import qutip
import numpy as np
from qutip.solver.sesolve import sesolve, SESolver
from qutip.solver.krylovsolve import krylovsolve
from qutip.solver.solver_base import Solver

# Deactivate warning for test without cython
from qutip.core.coefficient import WARN_MISSING_MODULE
WARN_MISSING_MODULE[0] = 0


all_ode_method = [
    method for method, integrator in SESolver.avail_integrators().items()
    if integrator.support_time_dependant
]

def _analytic(t, alpha):
    return ((1 - np.exp(-alpha * t)) / alpha)


class TestSeSolve():
    H0 = 0.2 * np.pi * qutip.sigmaz()
    H1 = np.pi * qutip.sigmax()
    tlist = np.linspace(0, 20, 200)
    args = {'alpha': 0.5}
    w_a = 0.35
    a = 0.5

    @pytest.mark.parametrize(['unitary_op'], [
        pytest.param(None, id="state"),
        pytest.param(qutip.qeye(2), id="unitary"),
    ])
    @pytest.mark.parametrize(['H', 'analytical'], [
        pytest.param(H1, lambda t, _: t, id='const_H'),
        pytest.param(lambda t, alpha: (np.pi * qutip.sigmax()
                                       * np.exp(-alpha * t)),
                     _analytic, id='func_H'),
        pytest.param([[H1, lambda t, args: np.exp(-args['alpha'] * t)]],
                     _analytic, id='list_func_H'),
        pytest.param([[H1, 'exp(-alpha*t)']],
                     _analytic, id='list_str_H'),
        pytest.param([[H1, np.exp(-args['alpha'] * tlist)]],
                     _analytic, id='list_array_H'),
        pytest.param(qutip.QobjEvo([[H1, 'exp(-alpha*t)']], args=args),
                     _analytic, id='QobjEvo_H'),
    ])
    def test_sesolve(self, H, analytical, unitary_op):
        """
        Compare integrated evolution with analytical result
        If U0 is not None then operator evo is checked
        Otherwise state evo
        """
        tol = 5e-3
        psi0 = qutip.basis(2, 0)
        options = {"progress_bar": None}

        if unitary_op is None:
            output = sesolve(H, psi0, self.tlist,
                             [qutip.sigmax(), qutip.sigmay(), qutip.sigmaz()],
                             args=self.args, options=options)
            sx, sy, sz = output.expect[0], output.expect[1], output.expect[2]
        else:
            output = sesolve(H, unitary_op, self.tlist, args=self.args,
                             options=options)
            sx = [qutip.expect(qutip.sigmax(), U * psi0)
                  for U in output.states]
            sy = [qutip.expect(qutip.sigmay(), U * psi0)
                  for U in output.states]
            sz = [qutip.expect(qutip.sigmaz(), U * psi0)
                  for U in output.states]

        sx_analytic = np.zeros(np.shape(self.tlist))
        sy_analytic = np.array(
            [-np.sin(2 * np.pi * analytical(t, self.args['alpha']))
             for t in self.tlist]
        )
        sz_analytic = np.array(
            [np.cos(2 * np.pi * analytical(t, self.args['alpha']))
             for t in self.tlist]
        )

        np.testing.assert_allclose(sx, sx_analytic, atol=tol)
        np.testing.assert_allclose(sy, sy_analytic, atol=tol)
        np.testing.assert_allclose(sz, sz_analytic, atol=tol)

    @pytest.mark.parametrize(['state_type'], [
        pytest.param("ket", id="ket"),
        pytest.param("unitary", id="unitary"),
    ])
    def test_sesolve_normalization(self, state_type):
        # non-hermitean H causes state to evolve non-unitarily
        H = qutip.Qobj([[1, -0.1j], [-0.1j, 1]])
        psi0 = qutip.basis(2, 0)
        options = {"normalize_output": True, "progress_bar": None}

        if state_type == "ket":
            output = sesolve(H, psi0, self.tlist, e_ops=[], options=options)
            norms = [state.norm() for state in output.states]
            np.testing.assert_allclose(
                norms, [1.0 for _ in self.tlist], atol=1e-15,
            )
        else:
            # evolution of unitaries should not be normalized
            U = qutip.qeye(2)
            output = sesolve(H, U, self.tlist, e_ops=[], options=options)
            norms = [state.norm() for state in output.states]
            assert all(norm > 2 for norm in norms[1:])

    @pytest.mark.parametrize(['unitary_op'], [
        pytest.param(None, id="state"),
        pytest.param(qutip.qeye(2), id="unitary"),
    ])
    @pytest.mark.parametrize('method', all_ode_method, ids=all_ode_method)
    def test_sesolve_method(self, method, unitary_op):
        """
        Compare integrated evolution with analytical result
        If U0 is not None then operator evo is checked
        Otherwise state evo
        """
        tol = 5e-3
        psi0 = qutip.basis(2, 0)
        options = {"method": method, "progress_bar": None}
        H = [[self.H1, 'exp(-alpha*t)']]

        if unitary_op is None:
            output = sesolve(H, psi0, self.tlist,
                             [qutip.sigmax(), qutip.sigmay(), qutip.sigmaz()],
                             args=self.args, options=options)
            sx, sy, sz = output.expect[0], output.expect[1], output.expect[2]
        else:
            output = sesolve(H, unitary_op, self.tlist,
                             args=self.args, options=options)
            sx = [qutip.expect(qutip.sigmax(), U * psi0)
                  for U in output.states]
            sy = [qutip.expect(qutip.sigmay(), U * psi0)
                  for U in output.states]
            sz = [qutip.expect(qutip.sigmaz(), U * psi0)
                  for U in output.states]

        sx_analytic = np.zeros(np.shape(self.tlist))
        sy_analytic = np.array(
            [np.sin(-2*np.pi * _analytic(t, self.args['alpha']))
             for t in self.tlist]
        )
        sz_analytic = np.array(
            [np.cos(2*np.pi *_analytic(t, self.args['alpha']))
             for t in self.tlist]
        )

        np.testing.assert_allclose(sx, sx_analytic, atol=tol)
        np.testing.assert_allclose(sy, sy_analytic, atol=tol)
        np.testing.assert_allclose(sz, sz_analytic, atol=tol)

    @pytest.mark.parametrize('normalize', [True, False],
                             ids=['Normalized', ''])
    @pytest.mark.parametrize(['H', 'args'],
        [pytest.param(H0 + H1, {}, id='const_H'),
         pytest.param(lambda t, a, w_a: (
             a * t * 0.2 * np.pi * qutip.sigmaz() +
             np.cos(w_a * t) * np.pi * qutip.sigmax()
         ), {'a':a, 'w_a':w_a}, id='func_H'),
         pytest.param([
             [H0, lambda t, args: args['a']*t],
             [H1, lambda t, args: np.cos(args['w_a']*t)]
         ], {'a':a, 'w_a':w_a}, id='list_func_H'),
         pytest.param([H0, [H1, 'cos(w_a*t)']], {'w_a':w_a}, id='list_str_H'),
    ])
    def test_compare_evolution(self, H, normalize, args, tol=5e-5):
        """
        Compare integrated evolution of unitary operator with state evo
        """
        psi0 = qutip.basis(2, 0)
        U0 = qutip.qeye(2)

        options = {
            "store_states": True,
            "normalize_output": normalize,
            "progress_bar": None
        }
        out_s = sesolve(H, psi0, self.tlist, [qutip.sigmax(),
                                              qutip.sigmay(),
                                              qutip.sigmaz()],
                        options=options, args=args)
        xs, ys, zs = out_s.expect[0], out_s.expect[1], out_s.expect[2]
        xss = [qutip.expect(qutip.sigmax(), U) for U in out_s.states]
        yss = [qutip.expect(qutip.sigmay(), U) for U in out_s.states]
        zss = [qutip.expect(qutip.sigmaz(), U) for U in out_s.states]

        np.testing.assert_allclose(xs, xss, atol=tol)
        np.testing.assert_allclose(ys, yss, atol=tol)
        np.testing.assert_allclose(zs, zss, atol=tol)

        if normalize:
            # propagator evolution is not normalized (yet?)
            tol = 5e-4
        out_u = sesolve(H, U0, self.tlist, options=options, args=args)
        xu = [qutip.expect(qutip.sigmax(), U * psi0) for U in out_u.states]
        yu = [qutip.expect(qutip.sigmay(), U * psi0) for U in out_u.states]
        zu = [qutip.expect(qutip.sigmaz(), U * psi0) for U in out_u.states]

        np.testing.assert_allclose(xs, xu, atol=tol)
        np.testing.assert_allclose(ys, yu, atol=tol)
        np.testing.assert_allclose(zs, zu, atol=tol)

    def test_sesolver_args(self):
        options = {"progress_bar": None}
        solver_obj = SESolver(qutip.QobjEvo([self.H0, [self.H1,'a']],
                                            args={'a': 1}),
                              options=options)
        res = solver_obj.run(qutip.basis(2,1), [0, 1, 2, 3],
                             e_ops=[qutip.num(2)], args={'a':0})
        np.testing.assert_allclose(res.expect[0], 1)

    def test_sesolver_pickling(self):
        e_ops = [qutip.sigmax(), qutip.sigmay(), qutip.sigmaz()]
        options = {"progress_bar": None}
        solver_obj = SESolver(self.H0 + self.H1,
                              options=options)
        solver_copy = pickle.loads(pickle.dumps(solver_obj))
        sx, sy, sz = solver_obj.run(qutip.basis(2,1), [0, 1, 2, 3],
                                    e_ops=e_ops).expect
        csx, csy, csz = solver_copy.run(qutip.basis(2,1), [0, 1, 2, 3],
                                       e_ops=e_ops).expect
        np.testing.assert_allclose(sx, csx)
        np.testing.assert_allclose(sy, csy)
        np.testing.assert_allclose(sz, csz)

    @pytest.mark.parametrize('method', all_ode_method, ids=all_ode_method)
    def test_sesolver_stepping(self, method):
        options = {
            "method": method,
            "atol": 1e-7,
            "rtol": 1e-8,
            "progress_bar": None
        }
        solver_obj = SESolver(
            qutip.QobjEvo([self.H1, lambda t, a: a], args={"a":0.25}),
            options=options
        )
        solver_obj.start(qutip.basis(2,0), 0)
        sr2 = -(2**.5)/2
        state = solver_obj.step(1)
        np.testing.assert_allclose(qutip.expect(qutip.sigmax(), state), 0.,
                                   atol=2e-6)
        np.testing.assert_allclose(qutip.expect(qutip.sigmay(), state), -1,
                                   atol=2e-6)
        np.testing.assert_allclose(qutip.expect(qutip.sigmaz(), state), 0.,
                                   atol=2e-6)

        state = solver_obj.step(2, args={"a":0.125})
        np.testing.assert_allclose(qutip.expect(qutip.sigmax(), state), 0.,
                                   atol=2e-6)
        np.testing.assert_allclose(qutip.expect(qutip.sigmay(), state), sr2,
                                   atol=2e-6)
        np.testing.assert_allclose(qutip.expect(qutip.sigmaz(), state), sr2,
                                   atol=2e-6)

        solver_obj.options = {
            "method": "adams",
            "atol": 1e-7,
            "rtol": 1e-8,
            "progress_bar": None
        }
        state = solver_obj.step(3, args={"a":0})
        np.testing.assert_allclose(qutip.expect(qutip.sigmax(), state), 0.,
                                   atol=2e-6)
        np.testing.assert_allclose(qutip.expect(qutip.sigmay(), state), sr2,
                                   atol=2e-6)
        np.testing.assert_allclose(qutip.expect(qutip.sigmaz(), state), sr2,
                                   atol=2e-6)


def test_sesolve_bad_H():
    with pytest.raises(TypeError):
        SESolver(np.eye(3))
    with pytest.raises(ValueError):
        SESolver(qutip.basis(3,1))


def test_sesolve_bad_state():
    solver = SESolver(qutip.qeye(4))
    with pytest.raises(TypeError):
        solver.start(qutip.basis(4,1).dag(), 0)
    with pytest.raises(TypeError):
        solver.start(qutip.basis(2,1) & qutip.basis(2,0), 0)


def test_sesolve_step_no_start():
    solver = SESolver(qutip.qeye(4))
    with pytest.raises(RuntimeError):
        solver.step(1)


@pytest.mark.parametrize("always_compute_step", [True, False])
def test_krylovsolve(always_compute_step):
    H = qutip.tensor([qutip.rand_herm(2) for _ in range(8)])
    psi0 = qutip.basis([2]*8, [1]*8)
    e_op = qutip.num(256)
    e_op.dims = H.dims
    tlist = np.linspace(0, 1, 11)
    ref = sesolve(H, psi0, tlist, e_ops=[e_op]).expect[0]
    options = {"always_compute_step", always_compute_step}
    krylov_sol = krylovsolve(H, psi0, tlist, 20, e_ops=[e_op]).expect[0]
    np.testing.assert_allclose(ref, krylov_sol)


def test_feedback():

    def f(t, A, qobj=None):
        return (A-2.)

    N = 4
    tol = 1e-14
    psi0 = qutip.basis(N, N-1)
    a = qutip.destroy(N)
    H = qutip.QobjEvo(
        [qutip.num(N), [a+a.dag(), f]],
        args={"A": SESolver.ExpectFeedback(qutip.num(N), default=3.)}
    )
    solver = qutip.SESolver(H)
    result = solver.run(psi0, np.linspace(0, 30, 301), e_ops=[qutip.num(N)])
    assert np.all(result.expect[0] > 2 - tol)



#cython: language_level=3
#cython: boundscheck=False, wraparound=False, initializedcheck=False

from libc.string cimport memset, memcpy
from libc.math cimport fabs
from libc.stdlib cimport abs
from libcpp.algorithm cimport lower_bound

import warnings
from qutip.settings import settings

cimport cython

from cpython cimport mem

import numpy as np
cimport numpy as cnp
from scipy.linalg cimport cython_blas as blas

from qutip.core.data.base import idxint_dtype
from qutip.core.data.base cimport idxint, Data
from qutip.core.data.dense cimport Dense
from qutip.core.data.csr cimport CSR
from qutip.core.data.dia cimport Dia
from qutip.core.data.tidyup cimport tidyup_dia
from qutip.core.data cimport csr, dense, dia
from qutip.core.data.add cimport iadd_dense, add_csr
from qutip.core.data.mul cimport imul_dense
from qutip.core.data.dense import OrderEfficiencyWarning

cnp.import_array()

cdef extern from *:
    void *PyMem_Calloc(size_t n, size_t elsize)

# This function is templated over integral types on import to allow `idxint` to
# be any signed integer (though likely things will only work for >=32-bit).  To
# change integral types, you only need to change the `idxint` definitions in
# `core.data.base` at compile-time.
cdef extern from "src/matmul_csr_vector.hpp" nogil:
    void _matmul_csr_vector[T](
        double complex *data, T *col_index, T *row_index,
        double complex *vec, double complex scale, double complex *out,
        T nrows)

cdef extern from "src/matmul_diag_vector.hpp" nogil:
    void _matmul_diag_vector[T](
        double complex *data, double complex *vec, double complex *out,
        T length, double complex scale)
    void _matmul_diag_block[T](
        double complex *data, double complex *vec, double complex *out,
        T length, T width)


__all__ = [
    'matmul', 'matmul_csr', 'matmul_dense', 'matmul_dia',
    'matmul_csr_dense_dense', 'matmul_dia_dense_dense', 'matmul_dense_dia_dense',
    'multiply', 'multiply_csr', 'multiply_dense', 'multiply_dia',
]


cdef int _check_shape(Data left, Data right, Data out=None) except -1 nogil:
    if left.shape[1] != right.shape[0]:
        raise ValueError(
            "incompatible matrix shapes "
            + str(left.shape)
            + " and "
            + str(right.shape)
        )
    if (
        out is not None
        and out.shape[0] != left.shape[0]
        and out.shape[1] != right.shape[1]
    ):
        raise ValueError(
            "incompatible output shape, got "
            + str(out.shape)
            + " but needed "
            + str((left.shape[0], right.shape[1]))
        )
    return 0

cdef idxint _matmul_csr_estimate_nnz(CSR left, CSR right):
    """
    Produce a sensible upper-bound for the number of non-zero elements that
    will be present in a matrix multiplication between the two matrices.
    """
    cdef idxint j, k, nnz=0
    cdef idxint ii, jj, kk
    cdef idxint nrows=left.shape[0], ncols=right.shape[1]
    # Setup mask array
    cdef idxint *mask = <idxint *> mem.PyMem_Malloc(ncols * sizeof(idxint))
    with nogil:
        for ii in range(ncols):
            mask[ii] = -1
        for ii in range(nrows):
            for jj in range(left.row_index[ii], left.row_index[ii+1]):
                j = left.col_index[jj]
                for kk in range(right.row_index[j], right.row_index[j+1]):
                    k = right.col_index[kk]
                    if mask[k] != ii:
                        mask[k] = ii
                        nnz += 1
    mem.PyMem_Free(mask)
    return nnz


cpdef CSR matmul_csr(CSR left, CSR right, double complex scale=1, CSR out=None):
    """
    Multiply two CSR matrices together to produce another CSR.  If `out` is
    specified, it must be pre-allocated with enough space to hold the output
    result.

    This is the operation
        ``out := left @ right``
    where `out` will be allocated if not supplied.

    Parameters
    ----------
    left : CSR
        CSR matrix on the left of the multiplication.
    right : CSR
        CSR matrix on the right of the multiplication.
    out : optional CSR
        Allocated space to store the result.  This must have enough space in
        the `data`, `col_index` and `row_index` pointers already allocated.

    Returns
    -------
    out : CSR
        The result of the matrix multiplication.  This will be the same object
        as the input parameter `out` if that was supplied.
    """
    _check_shape(left, right)
    cdef idxint nnz = _matmul_csr_estimate_nnz(left, right)
    if out is not None:
        raise TypeError("passing an `out` entry for CSR operations makes no sense")
    out = csr.empty(left.shape[0], right.shape[1], nnz if nnz != 0 else 1)
    if nnz == 0 or csr.nnz(left) == 0 or csr.nnz(right) == 0:
        # Ensure the out array row_index is zeroed.  The others need not be,
        # because they don't represent valid entries since row_index is zeroed.
        with nogil:
            memset(&out.row_index[0], 0, (out.shape[0] + 1) * sizeof(idxint))
        return out

    # Initialise actual matrix multiplication.
    nnz = 0
    cdef idxint head, length, row_l, ptr_l, row_r, ptr_r, col_r, tmp
    cdef idxint nrows=left.shape[0], ncols=right.shape[1]
    cdef double complex val
    cdef double complex *sums
    cdef idxint *nxt
    cdef double tol = 0
    if settings.core['auto_tidyup']:
        tol = settings.core['auto_tidyup_atol']
    sums = <double complex *> PyMem_Calloc(ncols, sizeof(double complex))
    nxt = <idxint *> mem.PyMem_Malloc(ncols * sizeof(idxint))
    with nogil:
        for col_r in range(ncols):
            nxt[col_r] = -1

        # Perform operation.
        out.row_index[0] = 0
        for row_l in range(nrows):
            head = -2
            length = 0
            for ptr_l in range(left.row_index[row_l], left.row_index[row_l+1]):
                row_r = left.col_index[ptr_l]
                val = left.data[ptr_l]
                for ptr_r in range(right.row_index[row_r], right.row_index[row_r+1]):
                    col_r = right.col_index[ptr_r]
                    sums[col_r] += val * right.data[ptr_r]
                    if nxt[col_r] == -1:
                        nxt[col_r] = head
                        head = col_r
                        length += 1
            for col_r in range(length):
                if fabs(sums[head].real) < tol:
                    sums[head].real = 0
                if fabs(sums[head].imag) < tol:
                    sums[head].imag = 0
                if sums[head] != 0:
                    out.col_index[nnz] = head
                    out.data[nnz] = scale * sums[head]
                    nnz += 1
                tmp = head
                head = nxt[head]
                nxt[tmp] = -1
                sums[tmp] = 0
            out.row_index[row_l + 1] = nnz
    mem.PyMem_Free(sums)
    mem.PyMem_Free(nxt)
    return out


cpdef Dense matmul_csr_dense_dense(CSR left, Dense right,
                                   double complex scale=1, Dense out=None):
    """
    Perform the operation
        ``out := scale * (left @ right) + out``
    where `left`, `right` and `out` are matrices.  `scale` is a complex scalar,
    defaulting to 1.

    If `out` is not given, it will be allocated as if it were a zero matrix.
    """
    _check_shape(left, right, out)
    cdef Dense tmp = None
    if out is None:
        out = dense.zeros(left.shape[0], right.shape[1], right.fortran)
    if bool(right.fortran) != bool(out.fortran):
        msg = (
            "out matrix is {}-ordered".format('Fortran' if out.fortran else 'C')
            + " but input is {}-ordered".format('Fortran' if right.fortran else 'C')
        )
        warnings.warn(msg, OrderEfficiencyWarning)
        # Rather than making loads of copies of the same code, we just moan at
        # the user and then transpose one of the arrays.  We prefer to have
        # `right` in Fortran-order for cache efficiency.
        if right.fortran:
            tmp = out
            out = out.reorder()
        else:
            right = right.reorder()
    cdef idxint row, ptr, idx_r, idx_out, nrows=left.shape[0], ncols=right.shape[1]
    cdef double complex val
    if right.fortran:
        idx_r = idx_out = 0
        for _ in range(ncols):
            _matmul_csr_vector(left.data, left.col_index, left.row_index,
                               right.data + idx_r,
                               scale,
                               out.data + idx_out,
                               nrows)
            idx_out += nrows
            idx_r += right.shape[0]
    else:
        for row in range(nrows):
            for ptr in range(left.row_index[row], left.row_index[row + 1]):
                val = scale * left.data[ptr]
                idx_out = row * ncols
                idx_r = left.col_index[ptr] * ncols
                for _ in range(ncols):
                    out.data[idx_out] += val * right.data[idx_r]
                    idx_out += 1
                    idx_r += 1
    if tmp is None:
        return out
    memcpy(tmp.data, out.data, ncols * nrows * sizeof(double complex))
    return tmp


cpdef Dense matmul_dense(Dense left, Dense right, double complex scale=1, Dense out=None):
    """
    Perform the operation
        ``out := scale * (left @ right) + out``
    where `left`, `right` and `out` are matrices.  `scale` is a complex scalar,
    defaulting to 1.

    If `out` is not given, it will be allocated as if it were a zero matrix.
    """
    _check_shape(left, right, out)
    cdef double complex out_scale
    # If not supplied, it's more efficient from a memory allocation perspective
    # to do the calculation as `a*A.B + 0*C` with arbitrary C.
    if out is None:
        out = dense.empty(left.shape[0], right.shape[1], right.fortran)
        out_scale = 0
    else:
        out_scale = 1
    cdef double complex *a
    cdef double complex *b
    cdef char transa, transb
    cdef int m, n, k=left.shape[1], lda, ldb
    if right.shape[1] == 1:
        # Matrix Vector product
        a, b = left.data, right.data
        if left.fortran:
            lda = left.shape[0]
            transa = b'n'
            m = left.shape[0]
            n = left.shape[1]
        else:
            lda = left.shape[1]
            transa = b't'
            m = left.shape[1]
            n = left.shape[0]
        ldb = 1
        blas.zgemv(&transa, &m , &n, &scale, a, &lda, b, &ldb,
                   &out_scale, out.data, &ldb)
        return out
    # We use the BLAS routine zgemm for every single call and pretend that
    # we're always supplying it with Fortran-ordered matrices, but to achieve
    # what we want, we use the property of matrix multiplication that
    #   A.B = (B'.A')'
    # where ' is the matrix transpose, and that interpreting a Fortran-ordered
    # matrix as a C-ordered one is equivalent to taking the transpose.  If
    # `right` is supplied in C-order, then from Fortran's perspective we
    # actually have `B'`, so to retrieve `B` should we want to use it, we set
    # `transb = b't'`.  What we set `transa` and `transb` to depends on if we
    # need to switch the input order, _not_ whether we actually need B'.
    #
    # In order to make the output correct, we ensure that we put A.B in if the
    # output is Fortran ordered, or B'.A' (note no final transpose) if not.
    # This is actually more flexible than `np.dot` which requires that the
    # output is C-ordered.
    if out.fortran:
        # Need to make A.B
        a, b = left.data, right.data
        m, n = left.shape[0], right.shape[1]
        lda = left.shape[0] if left.fortran else left.shape[1]
        transa = b'n' if left.fortran else b't'
        ldb = right.shape[0] if right.fortran else right.shape[1]
        transb = b'n' if right.fortran else b't'
    else:
        # Need to make B'.A'
        a, b = right.data, left.data
        m, n = right.shape[1], left.shape[0]
        lda = right.shape[0] if right.fortran else right.shape[1]
        transa = b't' if right.fortran else b'n'
        ldb = left.shape[0] if left.fortran else left.shape[1]
        transb = b't' if left.fortran else b'n'
    blas.zgemm(&transa, &transb, &m, &n, &k, &scale, a, &lda, b, &ldb,
               &out_scale, out.data, &m)
    return out


cpdef Dia matmul_dia(Dia left, Dia right, double complex scale=1):
    _check_shape(left, right, None)
    # We could probably do faster than this...
    npoffsets = np.unique(np.add.outer(left.as_scipy().offsets, right.as_scipy().offsets))
    npoffsets = npoffsets[np.logical_and(npoffsets > -left.shape[0], npoffsets < right.shape[1])]
    cdef idxint[:] offsets = npoffsets
    if len(npoffsets) == 0:
        return dia.zeros(left.shape[0], right.shape[1])
    cdef idxint *ptr = &offsets[0]
    cdef size_t num_diag = offsets.shape[0], diag_out, diag_left, diag_right
    cdef idxint start_left, end_left, start_out, end_out, start, end, col, off_out
    npdata = np.zeros((num_diag, right.shape[1]), dtype=complex)
    cdef double complex[:, ::1] data = npdata

    with nogil:
      for diag_left in range(left.num_diag):
        for diag_right in range(right.num_diag):
          off_out = left.offsets[diag_left] + right.offsets[diag_right]
          if off_out <= -left.shape[0] or off_out >= right.shape[1]:
            continue
          diag_out = <idxint> (lower_bound(ptr, ptr + num_diag, off_out) - ptr)

          start_left = max(0, left.offsets[diag_left]) + right.offsets[diag_right]
          start_right = max(0, right.offsets[diag_right])
          start_out = max(0, off_out)
          end_left = min(left.shape[1], left.shape[0] + left.offsets[diag_left]) + right.offsets[diag_right]
          end_right = min(right.shape[1], right.shape[0] + right.offsets[diag_right])
          end_out = min(right.shape[1], left.shape[0] + off_out)
          start = max(start_left, start_right, start_out)
          end = min(end_left, end_right, end_out)

          for col in range(start, end):
              data[diag_out, col] += (
                scale
                * left.data[diag_left * left.shape[1] + col - right.offsets[diag_right]]
                * right.data[diag_right * right.shape[1] + col]
              )
    return Dia((npdata, npoffsets), shape=(left.shape[0], right.shape[1]), copy=False)


cpdef Dense matmul_dia_dense_dense(Dia left, Dense right, double complex scale=1, Dense out=None):
    _check_shape(left, right, out)
    cdef Dense tmp
    if out is not None and scale == 1.:
        tmp = out
        out = None
    else:
        tmp = dense.zeros(left.shape[0], right.shape[1], right.fortran)

    cdef idxint start_left, end_left, start_out, end_out, length, i, start_right
    cdef idxint col, strideR_in, strideC_in, strideR_out, strideC_out
    cdef size_t diag

    with nogil:
      strideR_in = right.shape[1] if not right.fortran else 1
      strideC_in = right.shape[0] if right.fortran else 1
      strideR_out = tmp.shape[1] if not tmp.fortran else 1
      strideC_out = tmp.shape[0] if tmp.fortran else 1

      if (
        (left.shape[0] == left.shape[1])
        and (strideC_in == 1)
        and (strideC_out == 1)
      ):
          #Fast track for easy case
          for diag in range(left.num_diag):
              _matmul_diag_block(
                  right.data + max(0, left.offsets[diag]) * strideR_in,
                  left.data + diag * left.shape[1] + max(0, left.offsets[diag]),
                  tmp.data + max(0, -left.offsets[diag]) * strideR_out,
                  left.shape[1] - abs(left.offsets[diag]),
                  right.shape[1]
              )

      elif (strideR_in == 1) and (strideR_out == 1):
        for col in range(right.shape[1]):
          for diag in range(left.num_diag):
            start_left = max(0, left.offsets[diag])
            end_left = min(left.shape[1], left.shape[0] + left.offsets[diag])
            start_out = max(0, -left.offsets[diag])
            end_out = min(left.shape[0], left.shape[1] - left.offsets[diag])
            length = min(end_left - start_left, end_out - start_out)
            start_right = start_left + col * strideC_in
            start_left += diag * left.shape[1]
            start_out += col * strideC_out
            _matmul_diag_vector(
                left.data + start_left,
                right.data + start_right,
                tmp.data + start_out,
                length, 1.
            )

      else:
        for col in range(right.shape[1]):
          for diag in range(left.num_diag):
            start_left = max(0, left.offsets[diag])
            end_left = min(left.shape[1], left.shape[0] + left.offsets[diag])
            start_out = max(0, -left.offsets[diag])
            end_out = min(left.shape[0], left.shape[1] - left.offsets[diag])
            length = min(end_left - start_left, end_out - start_out)
            for i in range(length):
              tmp.data[(start_out + i) * strideR_out + col * strideC_out] += (
                left.data[diag * left.shape[1] + i + start_left]
                * right.data[(start_left + i) * strideR_in + col * strideC_in]
              )

    if out is None and scale == 1.:
        out = tmp
    elif out is None:
        imul_dense(tmp, scale)
        out = tmp
    else:
        iadd_dense(out, tmp, scale)

    return out


cpdef Dense matmul_dense_dia_dense(Dense left, Dia right, double complex scale=1, Dense out=None):
    _check_shape(left, right, out)
    cdef Dense tmp
    if out is not None and scale == 1.:
        tmp = out
        out = None
    else:
        tmp = dense.zeros(left.shape[0], right.shape[1], left.fortran)

    cdef idxint start_left, end_right, start_out, end_out, length, i, start_right
    cdef idxint row, strideR_in, strideC_in, strideR_out, strideC_out
    cdef size_t diag

    with nogil:
      strideR_in = left.shape[1] if not left.fortran else 1
      strideC_in = left.shape[0] if left.fortran else 1
      strideR_out = tmp.shape[1] if not tmp.fortran else 1
      strideC_out = tmp.shape[0] if tmp.fortran else 1

      if (
        (right.shape[0] == right.shape[1])
        and (strideR_in == 1)
        and (strideR_out == 1)
      ):
          #Fast track for easy case
          for diag in range(right.num_diag):
              _matmul_diag_block(
                  left.data + max(0, -right.offsets[diag]) * strideC_in,
                  right.data + diag * right.shape[1] + max(0, right.offsets[diag]),
                  tmp.data + max(0, right.offsets[diag]) * strideC_out,
                  right.shape[1] - abs(right.offsets[diag]),
                  left.shape[0]
              )

      elif (strideC_in == 1) and (strideC_out == 1):
        for row in range(left.shape[0]):
          for diag in range(right.num_diag):
            start_right = max(0, right.offsets[diag])
            end_right = min(right.shape[1], right.shape[0] + right.offsets[diag])
            start_out = max(0, right.offsets[diag])
            length = end_right - start_right
            start_left = max(0, -right.offsets[diag]) + row * strideR_in
            start_right += diag * right.shape[1]
            start_out = max(0, right.offsets[diag]) + row * strideR_out
            _matmul_diag_vector(
                right.data + start_right,
                left.data + start_left,
                tmp.data + start_out,
                length, 1.
            )

      else:
        for row in range(left.shape[0]):
          for diag in range(right.num_diag):
            start_right = max(0, right.offsets[diag])
            end_right = min(right.shape[1], right.shape[0] + right.offsets[diag])
            start_left = max(0, -right.offsets[diag])
            length = end_right - start_right
            for i in range(length):
              tmp.data[(start_right + i) * strideC_out + row * strideR_out] += (
                right.data[diag * right.shape[1] + i + start_right]
                * left.data[(start_left + i) * strideC_in + row * strideR_in]
              )

    if out is None and scale == 1.:
        out = tmp
    elif out is None:
        imul_dense(tmp, scale)
        out = tmp
    else:
        iadd_dense(out, tmp, scale)

    return out


cpdef CSR multiply_csr(CSR left, CSR right):
    """Element-wise multiplication of CSR matrices."""
    if left.shape[0] != right.shape[0] or left.shape[1] != right.shape[1]:
        raise ValueError(
            "incompatible matrix shapes "
            + str(left.shape)
            + " and "
            + str(right.shape)
        )
    cdef idxint col_left, left_nnz = csr.nnz(left)
    cdef idxint col_right, right_nnz = csr.nnz(right)
    cdef idxint ptr_left, ptr_right, ptr_left_max, ptr_right_max
    cdef idxint row, nnz=0, ncols=left.shape[1]
    cdef CSR out
    cdef list nans=[]
    # Fast paths for zero matrices.
    if right_nnz == 0 or left_nnz == 0:
        return csr.zeros(left.shape[0], left.shape[1])
    # Main path.
    out = csr.empty(left.shape[0], left.shape[1], max(left_nnz, right_nnz))
    out.row_index[0] = nnz
    ptr_left_max = ptr_right_max = 0

    for row in range(left.shape[0]):
        ptr_left = ptr_left_max
        ptr_left_max = left.row_index[row + 1]
        ptr_right = ptr_right_max
        ptr_right_max = right.row_index[row + 1]
        while ptr_left < ptr_left_max or ptr_right < ptr_right_max:
            col_left = left.col_index[ptr_left] if ptr_left < ptr_left_max else ncols + 1
            col_right = right.col_index[ptr_right] if ptr_right < ptr_right_max else ncols + 1
            if col_left == col_right:
                out.col_index[nnz] = col_left
                out.data[nnz] = left.data[ptr_left] * right.data[ptr_right]
                ptr_left += 1
                ptr_right += 1
                nnz += 1
            elif col_left <= col_right:
                if left.data[ptr_left] is np.nan:
                    # Test for NaN since `NaN * 0 = NaN`
                    nans.append((row, col_left))
                ptr_left += 1
            else:
                if right.data[ptr_right] != right.data[ptr_right]:
                    nans.append((row, col_right))
                ptr_right += 1
        out.row_index[row + 1] = nnz
    if nans:
        # We expect Nan to be rare enough that we don't allocate memory for
        # them, but add them here after the loop.
        nans_pos = np.array(nans, order='F', dtype=idxint_dtype)
        nnz = nans_pos.shape[0]
        nans_csr = csr.from_coo_pointers(
            <idxint *> cnp.PyArray_GETPTR2(nans_pos, 0, 0),
            <idxint *> cnp.PyArray_GETPTR2(nans_pos, 0, 1),
            <double complex *> cnp.PyArray_GETPTR1(
                np.array([np.nan]*nnz, dtype=np.complex128), 0),
            left.shape[0], left.shape[1], nnz
        )
        out = add_csr(out, nans_csr)
    return out


cpdef Dia multiply_dia(Dia left, Dia right):
    if left.shape[0] != right.shape[0] or left.shape[1] != right.shape[1]:
        raise ValueError(
            "incompatible matrix shapes "
            + str(left.shape)
            + " and "
            + str(right.shape)
        )
    cdef idxint diag_left=0, diag_right=0, out_diag=0, col
    cdef bint sorted=True
    cdef Dia out = dia.empty(left.shape[0], left.shape[1], min(left.num_diag, right.num_diag))

    with nogil:
      for diag_left in range(1, left.num_diag):
          if left.offsets[diag_left-1] > left.offsets[diag_left]:
              sorted = False
              continue
      if sorted:
          for diag_right in range(1, right.num_diag):
              if right.offsets[diag_right-1] > right.offsets[diag_right]:
                  sorted = False
                  continue

      if sorted:
        diag_left = 0
        diag_right = 0
        while diag_left < left.num_diag and diag_right < right.num_diag:
            if left.offsets[diag_left] == right.offsets[diag_right]:
                out.offsets[out_diag] = left.offsets[diag_left]
                for col in range(out.shape[1]):
                    if col >= left.shape[1] or col >= right.shape[1]:
                        out.data[out_diag * out.shape[1] + col] = 0
                    else:
                        out.data[out_diag * out.shape[1] + col] = (
                            left.data[diag_left * left.shape[1] + col] *
                            right.data[diag_right * right.shape[1] + col]
                        )
                out_diag += 1
                diag_left += 1
                diag_right += 1
            elif left.offsets[diag_left] < right.offsets[diag_right]:
                diag_left += 1
            else:
                diag_right += 1

      else:
        for diag_left in range(left.num_diag):
          for diag_right in range(right.num_diag):
            if left.offsets[diag_left] == right.offsets[diag_right]:
                out.offsets[out_diag] = left.offsets[diag_left]
                for col in range(right.shape[1]):
                    out.data[out_diag * out.shape[1] + col] = (
                        left.data[diag_left * left.shape[1] + col] *
                        right.data[diag_right * right.shape[1] + col]
                    )
                out_diag += 1
                break
      out.num_diag = out_diag

    if settings.core['auto_tidyup']:
        tidyup_dia(out, settings.core['auto_tidyup_atol'], True)
    return out


cpdef Dense multiply_dense(Dense left, Dense right):
    """Element-wise multiplication of Dense matrices."""
    if left.shape[0] != right.shape[0] or left.shape[1] != right.shape[1]:
        raise ValueError(
            "incompatible matrix shapes "
            + str(left.shape)
            + " and "
            + str(right.shape)
        )
    return Dense(left.as_ndarray() * right.as_ndarray(), copy=False)


from .dispatch import Dispatcher as _Dispatcher
import inspect as _inspect

# At the time of putting in the dispatchers, the idea of an "out" parameter
# isn't really supported in the model; since `out` would be potentially
# modified in-place, it couldn't safely go through a conversion process.  For
# the dispatched operation, then, we omit the `scale` and `out` parameters, and
# only dispatch on the operation `a @ b`.  If the `out` and `scale` parameters
# are needed, the library will have to manually do any relevant conversions,
# and then call a direct specialisation (which are exported to the `data`
# namespace).

matmul = _Dispatcher(
    _inspect.Signature([
        _inspect.Parameter('left', _inspect.Parameter.POSITIONAL_ONLY),
        _inspect.Parameter('right', _inspect.Parameter.POSITIONAL_ONLY),
        _inspect.Parameter('scale', _inspect.Parameter.POSITIONAL_OR_KEYWORD,
                           default=1),
    ]),
    name='matmul',
    module=__name__,
    inputs=('left', 'right'),
    out=True,
)
matmul.__doc__ =\
    """
    Compute the matrix multiplication of two matrices, with the operation
        scale * (left @ right)
    where `scale` is (optionally) a scalar, and `left` and `right` are
    matrices.

    Parameters
    ----------
    left : Data
        The left operand as either a bra or a ket matrix.

    right : Data
        The right operand as a ket matrix.

    scale : complex, optional
        The scalar to multiply the output by.
    """
matmul.add_specialisations([
    (CSR, CSR, CSR, matmul_csr),
    (CSR, Dense, Dense, matmul_csr_dense_dense),
    (Dense, Dense, Dense, matmul_dense),
    (Dia, Dia, Dia, matmul_dia),
    (Dia, Dense, Dense, matmul_dia_dense_dense),
    (Dense, Dia, Dense, matmul_dense_dia_dense),
], _defer=True)


multiply = _Dispatcher(
    _inspect.Signature([
        _inspect.Parameter('left', _inspect.Parameter.POSITIONAL_ONLY),
        _inspect.Parameter('right', _inspect.Parameter.POSITIONAL_ONLY),
    ]),
    name='multiply',
    module=__name__,
    inputs=('left', 'right'),
    out=True,
)
multiply.__doc__ =\
    """Element-wise multiplication of matrices."""
multiply.add_specialisations([
    (CSR, CSR, CSR, multiply_csr),
    (Dense, Dense, Dense, multiply_dense),
    (Dia, Dia, Dia, multiply_dia),
], _defer=True)


del _inspect, _Dispatcher


cdef Dense matmul_data_dense(Data left, Dense right):
    cdef Dense out
    if type(left) is CSR:
        out = matmul_csr_dense_dense(left, right)
    elif type(left) is Dense:
        out = matmul_dense(left, right)
    elif type(left) is Dia:
        out = matmul_dia_dense_dense(left, right)
    else:
        out = matmul(left, right)
    return out


cdef void imatmul_data_dense(Data left, Dense right, double complex scale, Dense out):
    if type(left) is CSR:
        matmul_csr_dense_dense(left, right, scale, out)
    elif type(left) is Dia:
        matmul_dia_dense_dense(left, right, scale, out)
    elif type(left) is Dense:
        matmul_dense(left, right, scale, out)
    else:
        iadd_dense(out, matmul(left, right, dtype=Dense), scale)



#cython: language_level=3

from qutip.core.data cimport CSR, Dense, Data, Dia

cpdef CSR imul_csr(CSR matrix, double complex value)
cpdef CSR mul_csr(CSR matrix, double complex value)
cpdef CSR neg_csr(CSR matrix)

cpdef Dense imul_dense(Dense matrix, double complex value)
cpdef Dense mul_dense(Dense matrix, double complex value)
cpdef Dense neg_dense(Dense matrix)

cpdef Dia imul_dia(Dia matrix, double complex value)
cpdef Dia mul_dia(Dia matrix, double complex value)
cpdef Dia neg_dia(Dia matrix)

cpdef Data imul_data(Data matrix, double complex value)



#cython: language_level=3
#cython: boundscheck=False, wraparound=False, initializedcheck=False

from qutip.core.data cimport CSR, Dense, Data, Dia

cpdef double complex expect_csr(CSR op, CSR state) except *
cpdef double complex expect_super_csr(CSR op, CSR state) except *

cpdef double complex expect_csr_dense(CSR op, Dense state) except *
cpdef double complex expect_super_csr_dense(CSR op, Dense state) except * nogil

cpdef double complex expect_dense(Dense op, Dense state) except *
cpdef double complex expect_super_dense(Dense op, Dense state) except * nogil

cpdef double complex expect_dia(Dia op, Dia state) except *
cpdef double complex expect_super_dia(Dia op, Dia state) except *

cpdef double complex expect_dia_dense(Dia op, Dense state) except *
cpdef double complex expect_super_dia_dense(Dia op, Dense state) except *

cdef double complex expect_data_dense(Data op, Dense state) except *
cdef double complex expect_super_data_dense(Data op, Dense state) except *
